import pytest
from finesse.script.exceptions import KatScriptError
from testutils.text import escape_full


@pytest.fixture
def fake_element_positional_args_cls(fake_element_cls):
    class FakeElementPositionalArgs(fake_element_cls):
        def __init__(self, name, a, b):
            super().__init__(name)
            self.a = a
            self.b = b

    return FakeElementPositionalArgs


@pytest.fixture
def fake_element_keyword_args_cls(fake_element_cls):
    class FakeElementKeywordArgs(fake_element_cls):
        def __init__(self, name, a=None, b=None):
            super().__init__(name)
            self.a = a
            self.b = b

    return FakeElementKeywordArgs


@pytest.fixture
def fake_element_keyword_args_suggestions_cls(fake_element_cls):
    class FakeElementKeywordArgsSuggestions(fake_element_cls):
        def __init__(self, name, apple=None, banana=None):
            super().__init__(name)
            self.apple = apple
            self.banana = banana

    return FakeElementKeywordArgsSuggestions


@pytest.fixture
def fake_element_mixed_args_cls(fake_element_cls):
    class FakeElementMixedArgs(fake_element_cls):
        def __init__(self, name, a, b, c, d=None):
            super().__init__(name)
            self.a = a
            self.b = b
            self.c = c
            self.d = d

    return FakeElementMixedArgs


@pytest.fixture
def fake_element_variadic_mixed_args_cls(fake_element_cls):
    class FakeElementVariadicMixedArgs(fake_element_cls):
        def __init__(self, name, a, b, c, *myargs, d=None, **mykwargs):
            super().__init__(name)
            self.a = a
            self.b = b
            self.c = c
            self.myargs = myargs
            self.d = d
            self.mykwargs = mykwargs

    return FakeElementVariadicMixedArgs


@pytest.fixture
def fake_command_positional_args_func():
    def fake_command_positional_args(model, a, b, c):
        pass

    return fake_command_positional_args


@pytest.fixture
def fake_command_keyword_args_func():
    def fake_command_keyword_args(model, a=None, b=None, c=None):
        pass

    return fake_command_keyword_args


@pytest.fixture
def fake_command_mixed_args_func():
    def fake_command_mixed_args(model, a, b, c, d=None):
        pass

    return fake_command_mixed_args


@pytest.fixture
def fake_command_variadic_mixed_args_func():
    def fake_command_variadic_mixed_args(model, a, b, c, *myargs, d=None, **mykwargs):
        pass

    return fake_command_variadic_mixed_args


@pytest.fixture
def spec(
    spec,
    set_spec_constructs,
    fake_element_adapter_factory,
    fake_element_positional_args_cls,
    fake_element_keyword_args_cls,
    fake_element_keyword_args_suggestions_cls,
    fake_element_mixed_args_cls,
    fake_element_variadic_mixed_args_cls,
    fake_command_adapter_factory,
    fake_command_positional_args_func,
    fake_command_keyword_args_func,
    fake_command_mixed_args_func,
    fake_command_variadic_mixed_args_func,
    finesse_binop_add,
    finesse_unop_neg,
):
    spec.register_element(
        fake_element_adapter_factory(
            fake_element_positional_args_cls, "fake_element_with_args"
        )
    )
    spec.register_element(
        fake_element_adapter_factory(
            fake_element_keyword_args_cls, "fake_element_with_kwargs"
        )
    )
    spec.register_element(
        fake_element_adapter_factory(
            fake_element_keyword_args_suggestions_cls,
            "fake_element_with_kwargs_suggestions",
        )
    )
    spec.register_element(
        fake_element_adapter_factory(fake_element_mixed_args_cls, "fake_element_mixed")
    )
    spec.register_element(
        fake_element_adapter_factory(
            fake_element_variadic_mixed_args_cls, "fake_element_variadic_mixed"
        )
    )
    spec.register_command(
        fake_command_adapter_factory(
            fake_command_positional_args_func, "fake_command_with_args"
        )
    )
    spec.register_command(
        fake_command_adapter_factory(
            fake_command_keyword_args_func, "fake_command_with_kwargs"
        )
    )
    spec.register_command(
        fake_command_adapter_factory(fake_command_mixed_args_func, "fake_command_mixed")
    )
    spec.register_command(
        fake_command_adapter_factory(
            fake_command_variadic_mixed_args_func, "fake_command_variadic_mixed"
        )
    )
    # Have to use real Finesse operator here because the builder matches against Finesse
    # operations.
    set_spec_constructs(
        "binary_operators",
        {"+": finesse_binop_add},
        "unary_operators",
        {"-": finesse_unop_neg},
    )

    return spec


@pytest.mark.parametrize(
    "script,error",
    (
        ## Elements
        # The error handler should compensate for the `self` argument of the setters
        # that are not exposed in KatScript. We also consider the `name` parameter not
        # to be an "argument" in KatScript, so that gets dealt with too.
        # Element with positional arguments.
        pytest.param(
            "fake_element_with_args myelement",
            "\nline 1: 'fake_element_with_args' missing 2 required positional arguments: 'a' and 'b'\n"
            "-->1: fake_element_with_args myelement\n"
            "                                       ^\n"
            "Syntax: fake_element_with_args name a b",
            id="two-arg-element-called-with-no-args",
        ),
        pytest.param(
            "fake_element_with_args myelement 1",
            "\nline 1: 'fake_element_with_args' missing 1 required positional argument: 'b'\n"
            "-->1: fake_element_with_args myelement 1\n"
            "                                         ^\n"
            "Syntax: fake_element_with_args name a b",
            id="two-arg-element-called-with-one-arg",
        ),
        # Element with positional and keyword arguments - no arguments.
        pytest.param(
            "fake_element_mixed myelement",
            "\nline 1: 'fake_element_mixed' missing 3 required positional arguments: 'a', 'b', and 'c'\n"
            "-->1: fake_element_mixed myelement\n"
            "                                   ^\n"
            "Syntax: fake_element_mixed name a b c d=none",
            id="four-mixed-arg-element-called-with-no-args",
        ),
        # Element with positional and keyword arguments - called by position.
        pytest.param(
            "fake_element_mixed myelement 1",
            "\nline 1: 'fake_element_mixed' missing 2 required positional arguments: 'b' and 'c'\n"
            "-->1: fake_element_mixed myelement 1\n"
            "                                     ^\n"
            "Syntax: fake_element_mixed name a b c d=none",
            id="four-mixed-arg-element-called-with-one-arg",
        ),
        # Element with positional and keyword arguments - called by keyword.
        pytest.param(
            "fake_element_mixed myelement a=1",
            "\nline 1: 'fake_element_mixed' missing 2 required positional arguments: 'b' and 'c'\n"
            "-->1: fake_element_mixed myelement a=1\n"
            "                                       ^\n"
            "Syntax: fake_element_mixed name a b c d=none",
            id="four-mixed-arg-element-called-with-one-kwarg",
        ),
        # Element with positional and keyword arguments - mixed call.
        pytest.param(
            "fake_element_mixed myelement 1 b=2",
            "\nline 1: 'fake_element_mixed' missing 1 required positional argument: 'c'\n"
            "-->1: fake_element_mixed myelement 1 b=2\n"
            "                                         ^\n"
            "Syntax: fake_element_mixed name a b c d=none",
            id="four-mixed-arg-element-called-with-one-arg-one-kwarg",
        ),
        # Element with variadic positional and keyword arguments - no arguments.
        pytest.param(
            "fake_element_variadic_mixed myelement",
            "\nline 1: 'fake_element_variadic_mixed' missing 3 required positional arguments: 'a', 'b', and 'c'\n"
            "-->1: fake_element_variadic_mixed myelement\n"
            "                                            ^\n"
            "Syntax: fake_element_variadic_mixed name a b c *myargs d=none **mykwargs",
            id="four-variadic-mixed-arg-element-called-with-no-args",
        ),
        # Element with variadic positional and keyword arguments - called by position.
        pytest.param(
            "fake_element_variadic_mixed myelement 1",
            "\nline 1: 'fake_element_variadic_mixed' missing 2 required positional arguments: 'b' and 'c'\n"
            "-->1: fake_element_variadic_mixed myelement 1\n"
            "                                              ^\n"
            "Syntax: fake_element_variadic_mixed name a b c *myargs d=none **mykwargs",
            id="four-variadic-mixed-arg-element-called-with-one-arg",
        ),
        # Element with variadic positional and keyword arguments - called by keyword.
        pytest.param(
            "fake_element_variadic_mixed myelement a=1",
            "\nline 1: 'fake_element_variadic_mixed' missing 2 required positional arguments: 'b' and 'c'\n"
            "-->1: fake_element_variadic_mixed myelement a=1\n"
            "                                                ^\n"
            "Syntax: fake_element_variadic_mixed name a b c *myargs d=none **mykwargs",
            id="four-variadic-mixed-arg-element-called-with-one-kwarg",
        ),
        # Element with variadic positional and keyword arguments - mixed call.
        pytest.param(
            "fake_element_variadic_mixed myelement 1 b=2",
            "\nline 1: 'fake_element_variadic_mixed' missing 1 required positional argument: 'c'\n"
            "-->1: fake_element_variadic_mixed myelement 1 b=2\n"
            "                                                  ^\n"
            "Syntax: fake_element_variadic_mixed name a b c *myargs d=none **mykwargs",
            id="four-variadic-mixed-arg-element-called-with-one-arg-one-kwarg",
        ),
        ## Commands
        # The error handler should compensate for the `self` (for direct setters) or
        # `model` (for setter proxies) of the setter arguments that are not exposed in
        # KatScript.
        # Command with positional arguments.
        pytest.param(
            "fake_command_with_args()",
            "\nline 1: 'fake_command_with_args' missing 3 required positional arguments: 'a', 'b', and 'c'\n"
            "-->1: fake_command_with_args()\n"
            "                             ^\n"
            "Syntax: fake_command_with_args(a, b, c)",
            id="three-arg-command-called-with-no-args",
        ),
        pytest.param(
            "fake_command_with_args(1)",
            "\nline 1: 'fake_command_with_args' missing 2 required positional arguments: 'b' and 'c'\n"
            "-->1: fake_command_with_args(1)\n"
            "                              ^\n"
            "Syntax: fake_command_with_args(a, b, c)",
            id="three-arg-command-called-with-one-arg",
        ),
        pytest.param(
            "fake_command_with_args(1, 2)",
            "\nline 1: 'fake_command_with_args' missing 1 required positional argument: 'c'\n"
            "-->1: fake_command_with_args(1, 2)\n"
            "                                 ^\n"
            "Syntax: fake_command_with_args(a, b, c)",
            id="three-arg-command-called-with-two-args",
        ),
        # Command with positional and keyword arguments - no arguments.
        pytest.param(
            "fake_command_mixed()",
            "\nline 1: 'fake_command_mixed' missing 3 required positional arguments: 'a', 'b', and 'c'\n"
            "-->1: fake_command_mixed()\n"
            "                         ^\n"
            "Syntax: fake_command_mixed(a, b, c, d=none)",
            id="four-mixed-arg-command-called-with-no-args",
        ),
        # Command with positional and keyword arguments - called by position.
        pytest.param(
            "fake_command_mixed(1)",
            "\nline 1: 'fake_command_mixed' missing 2 required positional arguments: 'b' and 'c'\n"
            "-->1: fake_command_mixed(1)\n"
            "                          ^\n"
            "Syntax: fake_command_mixed(a, b, c, d=none)",
            id="four-mixed-arg-command-called-with-one-arg",
        ),
        pytest.param(
            "fake_command_mixed(1, 2)",
            "\nline 1: 'fake_command_mixed' missing 1 required positional argument: 'c'\n"
            "-->1: fake_command_mixed(1, 2)\n"
            "                             ^\n"
            "Syntax: fake_command_mixed(a, b, c, d=none)",
            id="four-mixed-arg-command-called-with-two-args",
        ),
        # Command with positional and keyword arguments - called by keyword.
        pytest.param(
            "fake_command_mixed(a=1)",
            "\nline 1: 'fake_command_mixed' missing 2 required positional arguments: 'b' and 'c'\n"
            "-->1: fake_command_mixed(a=1)\n"
            "                            ^\n"
            "Syntax: fake_command_mixed(a, b, c, d=none)",
            id="four-mixed-arg-command-called-with-one-kwarg",
        ),
        pytest.param(
            "fake_command_mixed(a=1, b=2)",
            "\nline 1: 'fake_command_mixed' missing 1 required positional argument: 'c'\n"
            "-->1: fake_command_mixed(a=1, b=2)\n"
            "                                 ^\n"
            "Syntax: fake_command_mixed(a, b, c, d=none)",
            id="four-mixed-arg-command-called-with-two-kwargs",
        ),
        # Command with positional and keyword arguments - mixed call.
        pytest.param(
            "fake_command_mixed(1, b=2)",
            "\nline 1: 'fake_command_mixed' missing 1 required positional argument: 'c'\n"
            "-->1: fake_command_mixed(1, b=2)\n"
            "                               ^\n"
            "Syntax: fake_command_mixed(a, b, c, d=none)",
            id="four-mixed-arg-command-called-with-one-arg-one-kwarg",
        ),
        # Command with variadic positional and keyword arguments - no arguments.
        pytest.param(
            "fake_command_variadic_mixed()",
            "\nline 1: 'fake_command_variadic_mixed' missing 3 required positional arguments: 'a', 'b', and 'c'\n"
            "-->1: fake_command_variadic_mixed()\n"
            "                                  ^\n"
            "Syntax: fake_command_variadic_mixed(a, b, c, *myargs, d=none, **mykwargs)",
            id="four-variadic-mixed-arg-command-called-with-no-args",
        ),
        # Command with variadic positional and keyword arguments - called by position.
        pytest.param(
            "fake_command_variadic_mixed(1)",
            "\nline 1: 'fake_command_variadic_mixed' missing 2 required positional arguments: 'b' and 'c'\n"
            "-->1: fake_command_variadic_mixed(1)\n"
            "                                   ^\n"
            "Syntax: fake_command_variadic_mixed(a, b, c, *myargs, d=none, **mykwargs)",
            id="four-variadic-mixed-arg-command-called-with-one-arg",
        ),
        pytest.param(
            "fake_command_variadic_mixed(1, 2)",
            "\nline 1: 'fake_command_variadic_mixed' missing 1 required positional argument: 'c'\n"
            "-->1: fake_command_variadic_mixed(1, 2)\n"
            "                                      ^\n"
            "Syntax: fake_command_variadic_mixed(a, b, c, *myargs, d=none, **mykwargs)",
            id="four-variadic-mixed-arg-command-called-with-two-args",
        ),
        # Command with variadic positional and keyword arguments - called by keyword.
        pytest.param(
            "fake_command_variadic_mixed(a=1)",
            "\nline 1: 'fake_command_variadic_mixed' missing 2 required positional arguments: 'b' and 'c'\n"
            "-->1: fake_command_variadic_mixed(a=1)\n"
            "                                     ^\n"
            "Syntax: fake_command_variadic_mixed(a, b, c, *myargs, d=none, **mykwargs)",
            id="four-variadic-mixed-arg-command-called-with-one-kwarg",
        ),
        pytest.param(
            "fake_command_variadic_mixed(a=1, b=2)",
            "\nline 1: 'fake_command_variadic_mixed' missing 1 required positional argument: 'c'\n"
            "-->1: fake_command_variadic_mixed(a=1, b=2)\n"
            "                                          ^\n"
            "Syntax: fake_command_variadic_mixed(a, b, c, *myargs, d=none, **mykwargs)",
            id="four-variadic-mixed-arg-command-called-with-two-kwargs",
        ),
        # Command with variadic positional and keyword arguments - mixed call.
        pytest.param(
            "fake_command_variadic_mixed(1, b=2)",
            "\nline 1: 'fake_command_variadic_mixed' missing 1 required positional argument: 'c'\n"
            "-->1: fake_command_variadic_mixed(1, b=2)\n"
            "                                        ^\n"
            "Syntax: fake_command_variadic_mixed(a, b, c, *myargs, d=none, **mykwargs)",
            id="four-variadic-mixed-arg-command-called-with-one-arg-one-kwarg",
        ),
    ),
)
def test_not_enough_arguments(compiler, script, error):
    with pytest.raises(KatScriptError, match=escape_full(error)):
        compiler.compile(script)


@pytest.mark.parametrize(
    "script,error",
    (
        ## Elements
        # Element with positional arguments.
        pytest.param(
            "fake_element_with_args myelement 1 2 3",
            "\nline 1: 'fake_element_with_args' takes 2 positional arguments but 3 were given\n"
            "-->1: fake_element_with_args myelement 1 2 3\n"
            "                                           ^\n"
            "Syntax: fake_element_with_args name a b",
            id="two-arg-element-called-with-three-args",
        ),
        pytest.param(
            "fake_element_with_args myelement 1 2 3 4",
            "\nline 1: 'fake_element_with_args' takes 2 positional arguments but 4 were given\n"
            "-->1: fake_element_with_args myelement 1 2 3 4\n"
            "                                           ^ ^\n"
            "Syntax: fake_element_with_args name a b",
            id="two-arg-element-called-with-four-args",
        ),
        # Element with keyword arguments.
        pytest.param(
            "fake_element_with_kwargs myelement 1 2 3",
            "\nline 1: 'fake_element_with_kwargs' takes from 0 to 2 positional arguments but 3 were given\n"
            "-->1: fake_element_with_kwargs myelement 1 2 3\n"
            "                                             ^\n"
            "Syntax: fake_element_with_kwargs name a=none b=none",
            id="two-kwarg-element-called-with-three-args",
        ),
        pytest.param(
            "fake_element_with_kwargs myelement 1 2 3 4",
            "\nline 1: 'fake_element_with_kwargs' takes from 0 to 2 positional arguments but 4 were given\n"
            "-->1: fake_element_with_kwargs myelement 1 2 3 4\n"
            "                                             ^ ^\n"
            "Syntax: fake_element_with_kwargs name a=none b=none",
            id="two-kwarg-element-called-with-four-args",
        ),
        # Element with positional and keyword arguments - called by position.
        pytest.param(
            "fake_element_mixed myelement 1 2 3 4 5",
            "\nline 1: 'fake_element_mixed' takes from 3 to 4 positional arguments but 5 were given\n"
            "-->1: fake_element_mixed myelement 1 2 3 4 5\n"
            "                                           ^\n"
            "Syntax: fake_element_mixed name a b c d=none",
            id="four-mixed-arg-element-called-with-five-args",
        ),
        pytest.param(
            "fake_element_mixed myelement 1 2 3 4 5 6",
            "\nline 1: 'fake_element_mixed' takes from 3 to 4 positional arguments but 6 were given\n"
            "-->1: fake_element_mixed myelement 1 2 3 4 5 6\n"
            "                                           ^ ^\n"
            "Syntax: fake_element_mixed name a b c d=none",
            id="four-mixed-arg-element-called-with-six-args",
        ),
        # Element with positional and keyword arguments - called by keyword.
        pytest.param(
            "fake_element_mixed myelement a=1 b=2 c=3 d=4 e=5",
            "\nline 1: 'fake_element_mixed' got an unexpected keyword argument 'e'\n"
            "-->1: fake_element_mixed myelement a=1 b=2 c=3 d=4 e=5\n"
            "                                                   ^\n"
            "Syntax: fake_element_mixed name a b c d=none",
            id="four-mixed-arg-element-called-with-four-kwargs-one-invalid-kwarg",
        ),
        pytest.param(
            "fake_element_mixed myelement a=1 b=2 c=3 d=4 e=5 f=6",
            "\nline 1: 'fake_element_mixed' got an unexpected keyword argument 'e'\n"
            "-->1: fake_element_mixed myelement a=1 b=2 c=3 d=4 e=5 f=6\n"
            "                                                   ^\n"
            "Syntax: fake_element_mixed name a b c d=none",
            id="four-mixed-arg-element-called-with-four-kwargs-two-invalid-kwargs",
        ),
        # Element with positional and keyword arguments - mixed call.
        pytest.param(
            "fake_element_mixed myelement 1 2 3 d=4 e=5",
            "\nline 1: 'fake_element_mixed' got an unexpected keyword argument 'e'\n"
            "-->1: fake_element_mixed myelement 1 2 3 d=4 e=5\n"
            "                                             ^\n"
            "Syntax: fake_element_mixed name a b c d=none",
            id="four-mixed-arg-element-called-with-three-args-one-kwarg-one-invalid-kwarg",
        ),
        pytest.param(
            "fake_element_mixed myelement 1 2 3 d=4 e=5 f=6",
            "\nline 1: 'fake_element_mixed' got an unexpected keyword argument 'e'\n"
            "-->1: fake_element_mixed myelement 1 2 3 d=4 e=5 f=6\n"
            "                                             ^\n"
            "Syntax: fake_element_mixed name a b c d=none",
            id="four-mixed-arg-element-called-with-three-args-one-kwarg-two-invalid-kwargs",
        ),
        ## Commands
        # Command with positional arguments.
        pytest.param(
            "fake_command_with_args(1, 2, 3, 4)",
            "\nline 1: 'fake_command_with_args' takes 3 positional arguments but 4 were given\n"
            "-->1: fake_command_with_args(1, 2, 3, 4)\n"
            "                                      ^\n"
            "Syntax: fake_command_with_args(a, b, c)",
            id="three-arg-command-called-with-four-args",
        ),
        pytest.param(
            "fake_command_with_args(1, 2, 3, 4, 5)",
            "\nline 1: 'fake_command_with_args' takes 3 positional arguments but 5 were given\n"
            "-->1: fake_command_with_args(1, 2, 3, 4, 5)\n"
            "                                      ^  ^\n"
            "Syntax: fake_command_with_args(a, b, c)",
            id="three-arg-command-called-with-five-args",
        ),
        # Command with keyword arguments.
        pytest.param(
            "fake_command_with_kwargs(1, 2, 3, 4)",
            "\nline 1: 'fake_command_with_kwargs' takes from 0 to 3 positional arguments but 4 were given\n"
            "-->1: fake_command_with_kwargs(1, 2, 3, 4)\n"
            "                                        ^\n"
            "Syntax: fake_command_with_kwargs(a=none, b=none, c=none)",
            id="three-kwarg-command-called-with-four-args",
        ),
        pytest.param(
            "fake_command_with_kwargs(1, 2, 3, 4, 5)",
            "\nline 1: 'fake_command_with_kwargs' takes from 0 to 3 positional arguments but 5 were given\n"
            "-->1: fake_command_with_kwargs(1, 2, 3, 4, 5)\n"
            "                                        ^  ^\n"
            "Syntax: fake_command_with_kwargs(a=none, b=none, c=none)",
            id="three-kwarg-command-called-with-five-args",
        ),
        # Command with positional and keyword arguments - called by position.
        pytest.param(
            "fake_command_mixed(1, 2, 3, 4, 5)",
            "\nline 1: 'fake_command_mixed' takes from 3 to 4 positional arguments but 5 were given\n"
            "-->1: fake_command_mixed(1, 2, 3, 4, 5)\n"
            "                                     ^\n"
            "Syntax: fake_command_mixed(a, b, c, d=none)",
            id="four-mixed-arg-command-called-with-five-args",
        ),
        pytest.param(
            "fake_command_mixed(1, 2, 3, 4, 5, 6)",
            "\nline 1: 'fake_command_mixed' takes from 3 to 4 positional arguments but 6 were given\n"
            "-->1: fake_command_mixed(1, 2, 3, 4, 5, 6)\n"
            "                                     ^  ^\n"
            "Syntax: fake_command_mixed(a, b, c, d=none)",
            id="four-mixed-arg-command-called-with-six-args",
        ),
        # Command with positional and keyword arguments - called by keyword.
        pytest.param(
            "fake_command_mixed(a=1, b=2, c=3, d=4, e=5)",
            "\nline 1: 'fake_command_mixed' got an unexpected keyword argument 'e'\n"
            "-->1: fake_command_mixed(a=1, b=2, c=3, d=4, e=5)\n"
            "                                             ^\n"
            "Syntax: fake_command_mixed(a, b, c, d=none)",
            id="four-mixed-arg-command-called-with-four-kwargs-one-invalid-kwarg",
        ),
        pytest.param(
            "fake_command_mixed(a=1, b=2, c=3, d=4, e=5, f=6)",
            "\nline 1: 'fake_command_mixed' got an unexpected keyword argument 'e'\n"
            "-->1: fake_command_mixed(a=1, b=2, c=3, d=4, e=5, f=6)\n"
            "                                             ^\n"
            "Syntax: fake_command_mixed(a, b, c, d=none)",
            id="four-mixed-arg-command-called-with-four-kwargs-two-invalid-kwargs",
        ),
    ),
)
def test_too_many_arguments(compiler, script, error):
    with pytest.raises(KatScriptError, match=escape_full(error)):
        compiler.compile(script)


@pytest.mark.parametrize(
    "script,error",
    (
        pytest.param(
            "fake_command_with_args(1, 2, 3, a=1)",
            "\nline 1: 'fake_command_with_args' got multiple values for argument 'a'\n"
            "-->1: fake_command_with_args(1, 2, 3, a=1)\n"
            "                             ^          ^\n"
            "Syntax: fake_command_with_args(a, b, c)",
            id="three-arg-command-called-with-three-args-one-duplicate-kwarg",
        ),
        pytest.param(
            "fake_command_variadic_mixed(1, 2, 3, a=1)",
            "\nline 1: 'fake_command_variadic_mixed' got multiple values for argument 'a'\n"
            "-->1: fake_command_variadic_mixed(1, 2, 3, a=1)\n"
            "                                  ^          ^\n"
            "Syntax: fake_command_variadic_mixed(a, b, c, *myargs, d=none, **mykwargs)",
            id="variadic-mixed-command-called-with-three-args-one-duplicate-kwarg",
        ),
        # Duplicate keyword argument errors triggered before extraneous positional
        # arguments.
        pytest.param(
            "fake_command_with_args(1, 2, 3, 4, a=1)",
            "\nline 1: 'fake_command_with_args' got multiple values for argument 'a'\n"
            "-->1: fake_command_with_args(1, 2, 3, 4, a=1)\n"
            "                             ^             ^\n"
            "Syntax: fake_command_with_args(a, b, c)",
            id="three-arg-command-called-with-three-args-one-invalid-arg-one-duplicate-kwarg",
        ),
        pytest.param(
            "fake_command_with_args(1, 2, 3, 4, 5, a=1)",
            "\nline 1: 'fake_command_with_args' got multiple values for argument 'a'\n"
            "-->1: fake_command_with_args(1, 2, 3, 4, 5, a=1)\n"
            "                             ^                ^\n"
            "Syntax: fake_command_with_args(a, b, c)",
            id="three-arg-command-called-with-three-args-two-invalid-args-one-duplicate-kwarg",
        ),
        pytest.param(
            "fake_command_with_args(1, 2, 3, a=1, b=2)",
            "\nline 1: 'fake_command_with_args' got multiple values for argument 'a'\n"
            "-->1: fake_command_with_args(1, 2, 3, a=1, b=2)\n"
            "                             ^          ^\n"
            "Syntax: fake_command_with_args(a, b, c)",
            id="three-arg-command-called-with-three-args-two-duplicate-kwargs",
        ),
    ),
)
def test_duplicate_arguments(compiler, script, error):
    """Duplicate arguments are not allowed.

    Note: duplicate keyword arguments (i.e. multiple arguments specified with the same
    keyword) are tested in ../resolver/test_arguments.py as they are caught by the
    resolver, not builder.
    """
    with pytest.raises(KatScriptError, match=escape_full(error)):
        compiler.compile(script)


@pytest.mark.parametrize(
    "script,error",
    (
        pytest.param(
            "fake_element_with_kwargs_suggestions name aapple=1",
            "\nline 1: 'fake_element_with_kwargs_suggestions' got an unexpected keyword argument 'aapple'."
            " Did you mean 'apple'?\n"
            "-->1: fake_element_with_kwargs_suggestions name aapple=1\n"
            "                                                ^^^^^^\n"
            "Syntax: fake_element_with_kwargs_suggestions name apple=none banana=none",
            id="suggestion",
        ),
        pytest.param(
            "fake_element_with_kwargs_suggestions name cantaloupe=1",
            "\nline 1: 'fake_element_with_kwargs_suggestions' got an unexpected keyword argument 'cantaloupe'\n"
            "-->1: fake_element_with_kwargs_suggestions name cantaloupe=1\n"
            "                                                ^^^^^^^^^^\n"
            "Syntax: fake_element_with_kwargs_suggestions name apple=none banana=none",
            id="no-suggestion",
        ),
    ),
)
def test_suggestions(compiler, script, error):
    # with pytest.raises(KatScriptError, match=escape_full(error)):
    try:
        compiler.compile(script)
    except KatScriptError as e:
        result = str(e)
    print(error)
    print("###########################")
    print(result)
    assert error == result
    # for c1, c2 in zip(error, result):
    #     print(repr(c1), repr(c2), c1 == c2)



import finesse
import numpy as np
from finesse.gaussian import optimise_HG00_q
from finesse.knm.tools import make_scatter_matrix
import pytest


@pytest.fixture
def model():
    model = finesse.Model()
    model.modes("even", maxtem=4)
    return model


@pytest.mark.parametrize("z1x", [-10000, 10000])
@pytest.mark.parametrize("z2x", [-10000, 10000])
@pytest.mark.parametrize("z1y", [-10000, 10000])
@pytest.mark.parametrize("z2y", [-10000, 10000])
@pytest.mark.parametrize("w1x", [0.95, 1.05])
@pytest.mark.parametrize("w2x", [0.95, 1.05])
@pytest.mark.parametrize("w1y", [0.95, 1.05])
@pytest.mark.parametrize("w2y", [0.95, 1.05])
def test_optimise_HG00_q(model, z1x, w1x, z1y, w1y, z2x, w2x, z2y, w2y):
    q1x = finesse.BeamParam(w0=w1x, z=z1x)
    q1y = finesse.BeamParam(w0=w1y, z=z1y)
    q2x = finesse.BeamParam(w0=w2x, z=z2x)
    q2y = finesse.BeamParam(w0=w2y, z=z2y)
    kmat = make_scatter_matrix(
        "bayerhelms", q1x, q2x, q1y, q2y, 0, 0, select=model.homs
    )
    E = np.zeros(len(model.homs), dtype=complex)
    E[0] = 1
    E = kmat.data @ E  # new mode vector
    q3x, q3y = optimise_HG00_q(
        E, (q2x, q2y), model.homs, accuracy=1e-6, max_iterations=100
    )
    assert abs(q3x.w0 - q1x.w0) / abs(q1x.w0) < 0.001
    assert abs(q3x.z - q1x.z) / abs(q1x.z) < 0.001
    assert abs(q3y.w0 - q1y.w0) / abs(q1y.w0) < 0.001
    assert abs(q3y.z - q1y.z) / abs(q1y.z) < 0.001



from __future__ import annotations

import pytest
from finesse.script.compiler import KatCompiler
from finesse.script.spec import KatSpec
from finesse.script.adapter import ItemAdapter, ElementFactory, ElementSetter
from finesse.components.general import Connector
from finesse.components.node import NodeDirection, NodeType


def _resolve(value):
    if isinstance(value, list):
        for index, item in enumerate(value):
            value[index] = _resolve(item)
    else:
        try:
            value = value.eval()
        except AttributeError:
            pass

    return value


# Fixture for fuzzing tests. Hypothesis requires package scoped fixtures. Registers a
# special component that accepts a single argument of any type, which can be used to
# test parsing KatScript to Python objects.
@pytest.fixture(scope="package")
def fuzz_argument():
    class FuzzElement(Connector):
        """A fake element for fuzzing."""

        def __init__(self, name, value):
            super().__init__(name)
            self.value = value

            # Add some ports (required by the parent class).
            self._add_port("p1", NodeType.OPTICAL)
            self.p1._add_node("i", NodeDirection.INPUT)
            self.p1._add_node("o", NodeDirection.OUTPUT)

            self._add_port("p2", NodeType.OPTICAL)
            self.p2._add_node("i", NodeDirection.INPUT)
            self.p2._add_node("o", NodeDirection.OUTPUT)

    # Note: this breaks the "rule" not to modify the default KatSpec in the
    # /tests/script/katscript tree, but it's the simplest way to test parsing of
    # arbitrary arguments.
    spec = KatSpec()

    # Register fuzz element.
    spec.register_element(
        ItemAdapter(
            full_name="fuzz",
            factory=ElementFactory(item_type=FuzzElement),
            setter=ElementSetter(item_type=FuzzElement),
        )
    )

    compiler = KatCompiler(spec=spec)

    def _(argument_script):
        model = compiler.compile(f"fuzz el1 {argument_script}")
        value = model.el1.value

        return _resolve(value)

    return _


@pytest.fixture
def DIRECTIVES_REFERENCE() -> set[str]:
    return {
        "ABCD",
        "AstigmaticLens",
        "Amplifier",
        "AmplitudeDetector",
        "AntiSqueezing",
        "AstigmatismDetector",
        "BeamPropertyDetector",
        "BeamTrace",
        "Beamsplitter",
        "ButterFilter",
        "CCD",
        "CCDPixel",
        "CCDScanLine",
        "Cavity",
        "CavityPropertyDetector",
        "Change",
        "Cheby1Filter",
        "ClassicalNoise",
        "DCFields",
        "Debug",
        "DegreeOfFreedom",
        "DirectionalBeamsplitter",
        "Eigenmodes",
        "FieldCamera",
        "FieldDetector",
        "FieldPixel",
        "FieldScanLine",
        "For",
        "FreeMass",
        "FrequencyLoss",
        "FrequencyResponse",
        "FrequencyResponse2",
        "FrequencyResponse3",
        "FrequencyResponse4",
        "GetErrorSignals",
        "Gouy",
        "Isolator",
        "KnmDetector",
        "Laser",
        "Lens",
        "Lock",
        "MathDetector",
        "Maximize",
        "Minimize",
        "Mirror",
        "ModeMismatchDetector",
        "Modulator",
        "MotionDetector",
        "NoiseProjection",
        "Nothing",
        "Noxaxis",
        "Operator",
        "OpticalBandpassFilter",
        "OptimalQ",
        "OptimiseRFReadoutPhaseDC",
        "Parallel",
        "Pendulum",
        "Plot",
        "PowerDetector",
        "PowerDetectorDemod1",
        "PowerDetectorDemod2",
        "PrintModel",
        "PrintModelAttr",
        "Printer",
        "PropagateAstigmaticBeam",
        "PropagateBeam",
        "PseudoLockCavity",
        "PseudoLockDRFPMI",
        "QuantumNoiseDetector",
        "QuantumNoiseDetectorDemod1",
        "QuantumNoiseDetectorDemod2",
        "QuantumShotNoiseDetector",
        "QuantumShotNoiseDetectorDemod1",
        "QuantumShotNoiseDetectorDemod2",
        "ReadoutDC",
        "ReadoutRF",
        "RunLocks",
        "SaveMatrix",
        "SensingMatrixDC",
        "Series",
        "SetLockGains",
        "SignalGenerator",
        "Space",
        "Squeezer",
        "SuspensionZPK",
        "Sweep",
        "Telescope",
        "UpdateMaps",
        "Wire",
        "X2axis",
        "X3axis",
        "Xaxis",
        "ZPKFilter",
        "abcd",
        "ad",
        "amplifier",
        "amplitude_detector",
        "antisqueezing",
        "astigd",
        "beam_property_detector",
        "beam_trace",
        "beamsplitter",
        "bp",
        "bs",
        "butter",
        "cav",
        "cavity",
        "cavity_property_detector",
        "ccd",
        "ccdline",
        "ccdpx",
        "change",
        "cheby1",
        "cp",
        "dbs",
        "dc_fields",
        "debug",
        "degree_of_freedom",
        "directional_beamsplitter",
        "dof",
        "eigenmodes",
        "fcam",
        "fd",
        "field_detector",
        "filter_butter",
        "filter_cheby1",
        "filter_zpk",
        "fline",
        "floss",
        "for",
        "fpx",
        "free_mass",
        "freqresp",
        "frequency_response",
        "freqresp2",
        "frequency_response2",
        "freqresp3",
        "frequency_response3",
        "freqresp4",
        "frequency_response4",
        "fsig",
        "gauss",
        "get_error_signals",
        "gouy",
        "isol",
        "isolator",
        "knmd",
        "l",
        "lambda",
        "laser",
        "alens",
        "lens",
        "link",
        "lock",
        "m",
        "math_detector",
        "mathd",
        "maximize",
        "minimize",
        "mirror",
        "mmd",
        "mod",
        "modes",
        "modulator",
        "motion_detector",
        "noise",
        "noise_projection",
        "nothing",
        "noxaxis",
        "obp",
        "operator",
        "opt_rf_readout_phase",
        "optbp",
        "optical_bandpass",
        "optimal_q_detector",
        "parallel",
        "pd",
        "pd1",
        "pd2",
        "pendulum",
        "phase_config",
        "plot",
        "power_detector_dc",
        "power_detector_demod_1",
        "power_detector_demod_2",
        "print",
        "print_model",
        "print_model_attr",
        "propagate_beam",
        "propagate_beam_astig",
        "pseudo_lock_cavity",
        "pseudo_lock_drfpmi",
        "qnoised",
        "qnoised1",
        "qnoised2",
        "qshot",
        "qshot1",
        "qshot2",
        "quantum_noise_detector",
        "quantum_noise_detector_demod_1",
        "quantum_noise_detector_demod_2",
        "quantum_shot_noise_detector",
        "quantum_shot_noise_detector_demod_1",
        "quantum_shot_noise_detector_demod_2",
        "readout_dc",
        "readout_rf",
        "run_locks",
        "s",
        "save_matrix",
        "sensing_matrix_dc",
        "series",
        "set_lock_gains",
        "sgen",
        "signal_generator",
        "space",
        "sq",
        "squeezer",
        "sus_zpk",
        "suspension_zpk",
        "sweep",
        "tel",
        "telescope",
        "tem",
        "update_maps",
        "var",
        "variable",
        "w",
        "wire",
        "x2axis",
        "x3axis",
        "xaxis",
        "xd",
        "zpk",
        "TestPoint",
        "test_point",
        "test",
    }



"""Operator based Actions to extract operators and perform operator based analyes, such
as calculating eigenmodes."""

from more_itertools import pairwise
from scipy.sparse import diags
import numpy as np

from ...components import Cavity
from ...solutions import BaseSolution
from .base import Action


class EigenmodesSolution(BaseSolution):
    """Contains the result of an Eigenmodes action. The start node is defined by the
    Cavity starting point.

    Attributes
    ----------
    connections : tuple((Node, Node))
        Node connections used in the round trip propagator

    roundtrip_matrix : array
        Combined round trip matrix operator for the cavity

    matrices : list[array]
        A list of operators for each connection

    eigvalues, eigvectors : array, array
        Eigen values and vectors of the round trip matrix

    cavity_planewave_loss : float
        Round trip loss for a planewave

    homs : array_like
        Array of HOMs used in the model at the time this was computed
    """

    def loss(self, remove_planewave_loss=False):
        """Computes the round trip loss of all the eigenmodes of the cavity. Eigenmodes
        are ordered by loss. Lowest loss may not be the fundamental mode.

        Parameters
        ----------
        remove_planewave_loss : bool, optional
            Whether to remove the roundtrip loss a plane wave would experience
            to see the loss induced from HOM effects.

        Returns
        -------
        index : array_like
            Indicies of ordering for the eigvalues and eigvectors of this solution
        loss : array_like
            Roundtrip loss of modes
        """
        idx = np.argsort(1 - abs(self.eigvalues) ** 2)
        eigx_values = self.eigvalues[idx]
        loss = 1 - abs(eigx_values) ** 2
        if remove_planewave_loss:
            loss = loss - self.cavity_planewave_loss
        return idx, loss

    def plot_roundtrip_loss(self, remove_planewave_loss=False, ax=None, **kwargs):
        """Plots the roundtrip loss of the cavity for each eigenmode.

        Parameters
        ----------
        remove_planewave_loss : bool, optional
            If True, remove the loss a planewave would experience to just see the
            effects from higher order modes.

        ax : Matplotlib.Axis, optional
            The axis to plot on to, if `None` a new figure is made

        \\**kwargs
            Keyword arguments passed to matplotlib.pyplot.semilogy for styling trace
        """
        import matplotlib.pyplot as plt

        if ax is not None:
            plt.sca(ax)

        idx = np.argsort(1 - abs(self.eigvalues) ** 2)
        eigx_values = self.eigvalues[idx]
        loss = 1 - abs(eigx_values) ** 2
        if remove_planewave_loss:
            loss = loss - self.cavity_planewave_loss

        plt.semilogy(loss / 1e-6, **kwargs)
        plt.xlabel("Eigenvalue index")
        if not remove_planewave_loss:
            plt.ylabel("Roundtrip loss [ppm]")
        else:
            plt.ylabel("Loss excess from planewave [ppm]")

    def plot_phase(self, scale=None, ax=None, **kwargs):
        """Plots the eigenmode phases.

        Parameters
        ----------
        scale : float
            Scale of scatter point size

        ax : Matplotlib.Axis, optional
            The axis to plot on to, if `None` a new figure is made

        \\**kwargs
            Keyword arguments passed to matplotlib.pyplot.scatter for styling trace
        """
        import matplotlib.pyplot as plt

        if ax is not None:
            plt.sca(ax)

        idx = np.argsort(1 - abs(self.eigvalues) ** 2)
        eigx_values = self.eigvalues[idx]
        eigx_phase = np.angle(eigx_values)
        eigx_phase = eigx_phase - eigx_phase[0]
        plt.scatter(np.arange(eigx_values.size), eigx_phase, scale, **kwargs)
        plt.xlabel("Eigenvalue index")
        plt.ylabel("Eigenvalue phase [rad]")

    def plot_field(
        self,
        mode_idx,
        *,
        x=None,
        y=None,
        samples=100,
        scale=3,
        ax=None,
        colorbar=True,
        **kwargs,
    ):
        """Plots a 2D optical field for one of the eigenmodes.

        x and y dimensions can be specified if required, otherwise it
        will return an area of `scale` times the spot sizes. When `x` and
        `y` are provided `scale` and `samples` will not do anything.

        Parameters
        ----------
        mode_idx : int
            index of the mode to plot
        x, y : ndarray, optional
            Specify x and y coordinates to plot beam
        samples : int, optional
            Number of sample points to use in x and y
        scale : float, optional
            Number of sample points to use in x and y
        ax : Axis, optional
            A Matplotlib axis to put the image on. If None,
            a new figure will be made.
        colorbar : bool
            When True the colorbar will be added
        \\**kwargs
            Extra keyword arguments will be passed to the
            pcolormesh plotting function.
        """
        from ...plotting import plot_field

        plot_field(
            modes=self.homs,
            amplitudes=self.eigvectors[:, mode_idx],
            qs=self.q,
            x=x,
            y=y,
            samples=samples,
            scale=scale,
            ax=ax,
            colorbar=colorbar,
            **kwargs,
        )


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Eigenmodes(Action):
    """For a given Cavity defined in a model, this action will compute the roundtrip
    operator and calculate the eigen-values and -vectors of the cavity. This will not
    give correct solutions for coupled cavities as these need to include additional
    effects.

    This can be used to determine what modes combination of modes are resonating in a
    cavity and the required tuning to make that mode resonate.

    Parameters
    ----------
    cavity : str or :class:`.Cavity`
        cavity name or :class:`.Cavity` instance
    frequency : float
        Optical carrier or signal frequency to use for calculating the operators
    name : str, optional
        Name of the solution generated by this action
    """

    def __init__(self, cavity: Cavity, frequency, *, name="eigenmodes"):
        super().__init__(name)
        self.cavity = cavity
        self.frequency = frequency

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        use_signal = False
        sim = state.sim
        model = state.model
        cav = model.elements[
            self.cavity if isinstance(self.cavity, str) else self.cavity.name
        ]
        # Get the connections (node) forming this cavity
        nodes = cav.path.nodes
        # need to complete the loop by adding the first element
        # to the end again
        nodes.append(nodes[0])
        f_idx = None
        # find the right frequency index
        for freq in state.sim.carrier.optical_frequencies.frequencies:
            if freq.f == float(self.frequency):
                f_idx = freq.index
                break

        if state.sim.signal is not None:
            for freq in state.sim.signal.optical_frequencies.frequencies:
                if freq.f == float(self.frequency):
                    use_signal = True
                    f_idx = freq.index
                    break

        if f_idx is None:
            raise RuntimeError(
                f"Could not find an optical carrier frequency with a value of {self.frequency}Hz"
            )

        sol = EigenmodesSolution(self.name)
        sol.cavity_planewave_loss = cav.loss
        sol.homs = model.homs.copy()
        sol.q = cav.source.q
        sol.connections = tuple(
            (n1.full_name, n2.full_name) for n1, n2 in pairwise(nodes)
        )
        sol.roundtrip_matrix = None
        sol.matrices = []
        # update sim
        if sim.is_modal:
            sim.modal_update()
        sim.carrier.refill()

        if use_signal:
            sim.signal.refill()
            sim_to_use = sim.signal
        else:
            sim_to_use = sim.carrier

        for _ in pairwise(nodes):
            # if we have a 1D array it's just a diagonal matrix
            # so convert it to a sparse array for easy multiplying
            # later
            if sim_to_use.connections[_][f_idx].view.ndim == 1:
                M = diags(sim_to_use.connections[_][f_idx][:])
            else:
                M = sim_to_use.connections[_][f_idx].view.copy()
            # Keep reference to each coupling we come across
            sol.matrices.append(M)
            # Compute roundtrip matrix as we go
            if sol.roundtrip_matrix is None:
                sol.roundtrip_matrix = M
            else:
                sol.roundtrip_matrix = M @ sol.roundtrip_matrix

        # Find eigen values and vectors of roundtrip
        sol.eigvalues, sol.eigvectors = np.linalg.eig(sol.roundtrip_matrix)
        return sol


class OperatorSolution(BaseSolution):
    connections: tuple[tuple[str, str], ...]
    operator: np.ndarray
    """Contains solution to the Operator action. The main result is the `operator`
    attribute which describes the operator taking the field from start to end node.

    Attributes
    ----------
    connections : [(Node, Node)]
        A list of node pairs describing the connections
        traversed to compute this operator

    operator : ndarray(ndim=2, dtype=complex)
        The operator describing the propagation from start
        to end node.
    """
    connections: tuple[tuple[str, str], ...]
    operator: np.ndarray


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Operator(Action):
    """This action can be used to extract operators out from a simulation for external
    use. The operators are defined by a path in the network between two nodes (via some
    other if more direction is required).

    The `model.path` method can be used to test which nodes are traversed before using
    this to extract operators if needed.

    Parameters
    ----------
    start_node : str
        Start node name
    end_node : str
        End node name
    via : str, optional
        Via node to use to specify a path with multiple options
    frequency : float, optional
        Optical carrier or signal frequency to use for calculating the operators
    name : str, optional
        Name of the solution generated by this action
    """

    def __init__(self, start_node, end_node, via=None, frequency=0, *, name="operator"):
        super().__init__(name)
        self.start_node = start_node
        self.end_node = end_node
        self.via = via
        self.frequency = frequency

    def _requests(self, model, memo, first=True):
        memo["input_nodes"].append(model.get(self.start_node))
        memo["output_nodes"].append(model.get(self.end_node))
        if self.via is not None:
            memo["input_nodes"].append(model.get(self.via))

    def _do(self, state):
        use_signal = False
        sim = state.sim
        model = state.model
        key = id(self)
        f_idx = None
        sol = OperatorSolution(self.name)
        # Try and get data already computed for this action
        ws = state.action_workspaces.get(key, None)
        if ws:
            nodes = ws["nodes"]
            connections = ws["connections"]
            f_idx = ws["f_idx"]
        else:
            try:
                frequency = float(self.frequency)
            except TypeError:
                frequency = float(self.frequency.value)
            # find the right frequency index
            for freq in state.sim.carrier.optical_frequencies.frequencies:
                if freq.f == frequency:
                    f_idx = freq.index
                    break

            if state.sim.signal is not None:
                for freq in state.sim.signal.optical_frequencies.frequencies:
                    if freq.f == float(self.frequency):
                        use_signal = True
                        f_idx = freq.index
                        break

            if f_idx is None:
                raise RuntimeError(
                    f"Could not find an optical carrier or signal frequency with a value of {self.frequency}Hz"
                )
            nodes = model.path(self.start_node, self.end_node, via_node=self.via).nodes
            connections = tuple(
                (n1.full_name, n2.full_name) for n1, n2 in pairwise(nodes)
            )
            ws = {
                "nodes": nodes,
                "connections": connections,
                "f_idx": f_idx,
            }
            state.action_workspaces[key] = ws

        # update sim
        if sim.is_modal:
            sim.modal_update()
        sim.carrier.refill()

        if use_signal:
            sim.signal.refill()
            sim_to_use = sim.signal
        else:
            sim_to_use = sim.carrier

        sol.connections = connections
        sol.operator = np.eye(sim.model_settings.num_HOMs)
        for _ in pairwise(nodes):
            # if we have a 1D array it's just a diagonal matrix
            # so convert it to a sparse array for easy multiplying
            # later
            # TODO can have some option to select between different
            # frequencies or signal/carrier at some point
            Mview = sim_to_use.connections[_]

            if Mview.ndim == 2:
                # We have a frequency scattering matrix
                # TODO : not sure on user interface for getting
                # different frequency couplings yet, for now it's
                # just same freq in and out
                M = Mview[f_idx, f_idx]
            else:
                # no frequency coupling
                M = Mview[f_idx]

            if M.view.ndim == 1:
                # TODO can probably write something faster
                # than making a sparse diagonal matrix here
                sol.operator = diags(M.view) @ (-sol.operator)
            else:
                sol.operator = M.view @ (-sol.operator)

        return sol



import pytest
from finesse.symbols import simplification


# Override default element to take two model parameter arguments.
@pytest.fixture
def fake_element_cls(fake_element_cls, fake_float_parameter):
    @fake_float_parameter("a", "Fake Parameter A")
    @fake_float_parameter("b", "Fake Parameter B")
    class FakeElement(fake_element_cls):
        def __init__(self, name, a=None, b=None):
            super().__init__(name)
            self.a = a
            self.b = b

    return FakeElement


@pytest.fixture
def spec(
    spec,
    set_spec_constructs,
    fake_element_adapter_factory,
    fake_element_cls,
    finesse_binop_add,
    finesse_binop_sub,
    finesse_binop_mul,
):
    spec.register_element(fake_element_adapter_factory(fake_element_cls))
    set_spec_constructs(
        "binary_operators",
        {"+": finesse_binop_add, "-": finesse_binop_sub, "*": finesse_binop_mul},
    )

    return spec


def test_reference_unsimplified(unbuilder, model, element_dump, fake_element_cls):
    model.add(fake_element_cls("myel1"))
    model.add(fake_element_cls("myel2"))
    model.myel1.b.value = 0
    model.myel1.a = 1 - model.myel1.b.ref
    model.myel2.a = model.myel1.a.ref
    model.myel2.b = model.myel1.b.ref
    dumps = iter(element_dump("fake_element", fake_element_cls, model))
    assert unbuilder.unbuild(next(dumps)) == "fake_element myel1 a=(1-myel1.b) b=0.0"
    assert unbuilder.unbuild(next(dumps)) == "fake_element myel2 a=myel1.a b=myel1.b"


def test_reference_simplified(unbuilder, model, element_dump, fake_element_cls):
    with simplification():
        model.add(fake_element_cls("myel1"))
        model.add(fake_element_cls("myel2"))
        model.myel1.b.value = 0
        model.myel1.a = 1 - model.myel1.b.ref
        model.myel2.a = model.myel1.a.ref
        model.myel2.b = model.myel1.b.ref
        dumps = iter(element_dump("fake_element", fake_element_cls, model))
        assert (
            unbuilder.unbuild(next(dumps))
            == "fake_element myel1 a=((-1*myel1.b)+1) b=0.0"
        )
        assert (
            unbuilder.unbuild(next(dumps)) == "fake_element myel2 a=myel1.a b=myel1.b"
        )



import pytest

from finesse import Model
from finesse.components import Laser, Beamsplitter, Mirror
from finesse.detectors import PowerDetector
from finesse.utilities.network_filter import (
    NetworkFilterBase,
    NetworkType,
    ComponentNetworkFilter,
    OpticalNetworkFilter,
    FullNetworkFilter,
)

from testutils.diff import assert_graphs_match


@pytest.fixture
def michelson():
    m = Model()
    m.link(Laser("l1"), Beamsplitter("bs1", R=0.5, T=0.5))
    m.link(m.bs1.p2, Mirror("ITMy", R=0.5, T=0.5), Mirror("ETMy", R=0.5, T=0.5))
    m.link(m.bs1.p3, Mirror("ITMx", R=0.5, T=0.5), Mirror("ETMx", R=0.5, T=0.5))
    m.add(PowerDetector("out", m.bs1.p4.o))
    return m


def test_network_type(michelson):
    assert_graphs_match(
        NetworkType.COMPONENT.get_network(michelson), michelson.to_component_network()
    )
    assert_graphs_match(
        NetworkType.OPTICAL.get_network(michelson), michelson.optical_network
    )
    assert_graphs_match(NetworkType.FULL.get_network(michelson), michelson.network)


@pytest.mark.parametrize(
    "network_type, expected",
    (
        (NetworkType.COMPONENT, ComponentNetworkFilter),
        (NetworkType.OPTICAL, OpticalNetworkFilter),
        (NetworkType.FULL, FullNetworkFilter),
    ),
)
def test_from_network_type(network_type, expected):
    assert network_type.filter_class is expected


def test_root_str(michelson):
    for root, expected in ((None, "l1"), ("l1", "l1"), (michelson.l1, "l1")):
        component_filter = ComponentNetworkFilter(michelson, root=root)
        assert component_filter.root_str == expected


@pytest.mark.parametrize(
    "filter_class, root",
    (
        (ComponentNetworkFilter, "l1.p1"),
        (OpticalNetworkFilter, "l1"),
        (FullNetworkFilter, "foo"),
    ),
)
def test_root_str_missing(michelson, filter_class, root):
    with pytest.raises(KeyError):
        filter_class(michelson, root=root)


def test_root_type_error(michelson):
    with pytest.raises(TypeError):
        ComponentNetworkFilter(michelson, root=michelson.l1.p1)
    with pytest.raises(TypeError):
        FullNetworkFilter(michelson, root=michelson.l1)


def test_negative_radius(michelson):
    with pytest.raises(ValueError):
        ComponentNetworkFilter(michelson, radius=-1).run()
    with pytest.raises(ValueError):
        ComponentNetworkFilter(michelson, radius=0).run()


@pytest.mark.parametrize(
    "filter_klass, root, radius, add_detectors, undirected, expected_nodes",
    (
        (ComponentNetworkFilter, "l1", 1, False, True, ("l1", "bs1")),
        (ComponentNetworkFilter, "l1", 2, False, True, ("l1", "bs1", "ITMy", "ITMx")),
        (
            ComponentNetworkFilter,
            "bs1",
            1,
            True,
            True,
            ("l1", "bs1", "ITMy", "ITMx", "out"),
        ),
        (OpticalNetworkFilter, "l1.p1.o", 1, False, True, ("l1.p1.o", "bs1.p1.i")),
        (OpticalNetworkFilter, "l1.p1.o", 1, False, False, ("l1.p1.o", "bs1.p1.i")),
        (
            OpticalNetworkFilter,
            "bs1.p2.o",
            1,
            False,
            False,
            ("bs1.p2.o", "ITMy.p1.i"),
        ),
        (
            OpticalNetworkFilter,
            "bs1.p2.o",
            1,
            False,
            True,
            ("bs1.p2.o", "ITMy.p1.i", "bs1.p1.i", "bs1.p4.i"),
        ),
    ),
)
def test_filter(
    michelson,
    filter_klass: type[NetworkFilterBase],
    root,
    radius,
    add_detectors,
    undirected,
    expected_nodes,
):
    result = filter_klass(
        michelson,
        root=root,
        radius=radius,
        add_detectors=add_detectors,
        undirected=undirected,
    ).run()
    assert set(expected_nodes) == set(result.nodes)



from finesse.components.workspace cimport ConnectorWorkspace, FillFuncWrapper
from finesse.element_workspace cimport BaseCValues

cdef class SignalGeneratorValues(BaseCValues):
    cdef public:
        double amplitude
        double phase

cdef class SignalGeneratorWorkspace(ConnectorWorkspace):
    cdef public:
        SignalGeneratorValues v
        Py_ssize_t rhs_idx
        double scaling

cdef object c_siggen_fill_rhs(ConnectorWorkspace cws)



"""Space-type objects representing physical distances between components."""

import logging
import types
import finesse
import numpy as np

from finesse.symbols import Variable, Matrix
from finesse.env import warn
from finesse.parameter import float_parameter
from finesse.exceptions import ContextualTypeError, ContextualValueError
from finesse.components.general import (
    Connector,
    InteractionType,
    borrows_nodes,
    LocalDegreeOfFreedom,
)
from finesse.components.readout import _Readout
from finesse.components.workspace import ConnectionSetting
from finesse.components.node import NodeType, Port, NodeDirection
from finesse.tracing import abcd

LOGGER = logging.getLogger(__name__)


@borrows_nodes()
@float_parameter(
    "L",
    "Length",
    validate="_check_L",
    units="m",
    is_geometric=True,
)
@float_parameter(
    "nr",
    "Refractive index",
    validate="_check_nr",
    is_geometric=True,
    changeable_during_simulation=False,
)
@float_parameter("user_gouy_x", "Gouy phase (x)", units="Degrees")
@float_parameter("user_gouy_y", "Gouy phase (y)", units="Degrees")
# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Space(Connector):
    """Represents a space between two components in the interferometer configuration,
    with a given length and index of refraction.

    There can be many spaces in a model which are not of much interest and do not
    need to be referred to. For example, the ``link`` command will create spaces
    between components without giving an explicit name, just specifying a distance
    between them.

    All Space elements are added to the Model's ``.spaces`` namespace so they can
    all easily be iterated over. You can also find spaces which had no name
    specified and thus have an autogenerated name. User specified Space's with
    names will be put in the main Model namespace so that they can be accessed
    like any other element in the model.

    Parameters
    ----------
    name : str, optional
        Name of newly created space. If not specified, a name is automatically
        generated.

    portA, portB : :class:`.Port`
        Ports to connect.

    L : float, optional
        Geometric length of newly created :class:`.Space` instance; defaults to 0.

    nr : float, optional
        Index of refraction of newly created :class:`.Space` instance; defaults to 1.0.

    user_gouy_x, user_gouy_y : float, optional
        User-defined gouy phase to override the calculated value.
    """

    def __init__(
        self,
        name,
        portA,
        portB,
        L=0.0,
        nr=1.0,
        user_gouy_x=None,
        user_gouy_y=None,
    ):
        # Override None name with auto https://gitlab.com/ifosim/finesse/finesse3/-/merge_requests/251
        if (portA is None) != (portB is None):
            warn(
                "Can't construct a space with only one port connected; ignoring ports."
            )
            portA = None
            portB = None

        if portA is not None and not isinstance(portA, Port):
            raise ContextualTypeError("portA", portA, allowed_types=(Port,), name=name)
        if portA is not None and not isinstance(portB, Port):
            raise ContextualTypeError("portB", portB, allowed_types=(Port,), name=name)

        if portA is not None and portA.type != NodeType.OPTICAL:
            raise ContextualValueError(
                {"portA": portA},
                f"must be an optical port for {name}.portA",
            )
        if portB is not None and portB.type != NodeType.OPTICAL:
            raise ContextualValueError(
                {"portB": portB},
                f"must be an optical port for {name}.portB",
            )

        if portA is not None and portB is not None:
            if portA is portB:
                raise ValueError("Port A and B are the same port.")
            for n1 in portA.nodes:
                for n2 in portB.nodes:
                    if n1 is n2:
                        raise ValueError(f"Port A and B contain the same node {n1}")

            for port in (portA, portB):
                if isinstance(port.component, _Readout):
                    # check if readout has copied over nodes from another port
                    port.component.check_connection()

            if portA.component._model is not portB.component._model:
                raise ValueError("Port A and B are not part of the same model")

        if name is None:
            auto_generated = True
            if portA is not None and portB is not None:
                compA = portA.component.name
                compB = portB.component.name
                name = f"{compA}_{portA.name}__{compB}_{portB.name}"
            else:
                raise ValueError(
                    "Cannot create an unconnected space without providing a name"
                )
            self._auto_generated = auto_generated
            self._auto_generated_name = name
        else:
            self._auto_generated = False
            self._auto_generated_name = None

        if self._auto_generated:
            super().__init__(self._auto_generated_name)
            self._namespace = (".spaces",)
        else:
            super().__init__(name)
            # Also put into main namespace if it has a specific name
            self._namespace = (".", ".spaces")

        self.__portA = portA
        self.__portB = portB
        self._add_to_model_namespace = True

        self.L = L
        self.nr = nr
        self.user_gouy_x = user_gouy_x
        self.user_gouy_y = user_gouy_y

        self._add_port("p1", NodeType.OPTICAL)
        self._add_port("p2", NodeType.OPTICAL)

        # Phase modulation input
        self._add_port("phs", NodeType.ELECTRICAL)
        self.phs._add_node("i", NodeDirection.INPUT)
        # Amplitude modulation input
        self._add_port("amp", NodeType.ELECTRICAL)
        self.amp._add_node("i", NodeDirection.INPUT)
        # strain input
        self._add_port("h", NodeType.ELECTRICAL)
        self.h._add_node("i", NodeDirection.INPUT)

        if portA is not None and portB is not None:
            self.connect(portA, portB)

        # Define typical degrees of freedom for this component
        self.dofs = types.SimpleNamespace()
        # Strain doesn't have a DC term really in Finesse
        # changing a space length won't generate a signal
        self.dofs.h = LocalDegreeOfFreedom(f"{self.name}.dofs.h", None, self.h.i, 1)

    def optical_equations(self):
        """Calculates the optical equations for the space component.

        Returns
        -------
            dict: A dictionary containing the optical equations for the space component.
                The keys match the optical couplings defined by the Space, and the values
                represent the corresponding complex coupling equations.
        """
        with finesse.symbols.simplification():
            settings = self._model._settings
            _f_ = Variable("_f_")
            _c0_ = finesse.symbols.CONSTANTS["c0"]
            pi = finesse.symbols.CONSTANTS["pi"]
            gouy = Matrix(f"{self.name}.gouy")

            pre_factor = 2 * pi * self.nr.ref * self.L.ref / _c0_
            phi = pre_factor * _f_

            if settings.is_modal:
                return {
                    f"{self.name}.P1i_P2o": np.exp(-1j * phi) * gouy,
                    f"{self.name}.P2i_P1o": np.exp(-1j * phi) * gouy,
                }
            else:
                # Otherwise we just get the phase from the propagation along the
                # space if the frequency is different to that of the model reference
                # wavelength
                return {
                    f"{self.name}.P1i_P2o": np.exp(-1j * phi),
                    f"{self.name}.P2i_P1o": np.exp(-1j * phi),
                }

    @property
    def gouy_x(self):
        # return user value if it is set
        # otherwise try to return calculated value or default

        if self.user_gouy_x.value is not None:
            return self.user_gouy_x.value
        else:
            try:
                return self.gouy(self.portA.o.qx, self.portB.i.qx)
            except RuntimeError:
                # No simulation, use the default value
                return 0.0

    @property
    def gouy_y(self):
        # return user value if it is set
        # otherwise try to return calculated value or default
        if self.user_gouy_y.value is not None:
            return self.user_gouy_y.value
        else:
            try:
                return self.gouy(self.portA.o.qy, self.portB.i.qy)
            except RuntimeError:
                # No simulation, use the default value
                return 0.0

    def gouy(self, q1, q2):
        """Computes the Gouy phase in degrees from beam parameters.

        Parameters
        ----------
        q1 : complex, :class:`.BeamParam`
            Starting beam parameter

        q2 : complex, :class:`.BeamParam`
            Ending beam parameter

        Returns
        -------
        out : float
            Gouy phase (in degrees)
        """
        return np.degrees(
            abs(np.arctan2(q1.real, q1.imag) - np.arctan2(q2.real, q2.imag))
        )

    @property
    def portA(self):
        return self.__portA

    @property
    def portB(self):
        return self.__portB

    def connect(self, portA, portB):
        """Sets the ports of this `Space`.

        Parameters
        ----------
        portA : :class:`.Port`, optional
            Port to connect

        portB : :class:`.Port`, optional
            Port to connect
        """
        if portA.is_connected:
            raise Exception(f"Port {portA} has already been connected to")
        if portB.is_connected:
            raise Exception(f"Port {portB} has already been connected to")

        # From the Space's perspective the input and output
        # nodes are swapped around for its ports
        self.p1._add_node("i", None, portA.o)
        self.p1._add_node("o", None, portA.i)
        self.p2._add_node("i", None, portB.o)
        self.p2._add_node("o", None, portB.i)

        self._register_node_coupling(
            "P1i_P2o",
            self.p1.i,
            self.p2.o,
            interaction_type=InteractionType.TRANSMISSION,
        )
        self._register_node_coupling(
            "P2i_P1o",
            self.p2.i,
            self.p1.o,
            interaction_type=InteractionType.TRANSMISSION,
        )
        self._register_node_coupling("SIGPHS_P1o", self.phs.i, self.p1.o)
        self._register_node_coupling("SIGPHS_P2o", self.phs.i, self.p2.o)
        self._register_node_coupling("SIGAMP_P1o", self.amp.i, self.p1.o)
        self._register_node_coupling("SIGAMP_P2o", self.amp.i, self.p2.o)
        self._register_node_coupling("H_P1o", self.h.i, self.p1.o)
        self._register_node_coupling("H_P2o", self.h.i, self.p2.o)

    def _get_workspace(self, sim):
        from finesse.components.modal.space import (
            space_carrier_fill,
            space_signal_fill,
            space_set_gouy,
            SpaceWorkspace,
        )

        _, is_changing = self._eval_parameters()

        carrier_refill = sim.carrier.any_frequencies_changing and (
            float(self.L.value) > 0 or self.L.is_changing
        )
        carrier_refill |= len(is_changing)
        carrier_refill |= self in sim.trace_forest
        carrier_refill |= self.user_gouy_x.is_changing
        carrier_refill |= self.user_gouy_y.is_changing

        ws = SpaceWorkspace(self, sim)

        # use user gouy if it is defined
        ws.use_user_gouy_x = self.user_gouy_x.value is not None
        ws.use_user_gouy_y = self.user_gouy_y.value is not None

        ws.set_gouy_function(space_set_gouy)
        # Set the fill function for this simulation
        ws.carrier.add_fill_function(space_carrier_fill, carrier_refill)
        ws.carrier.connection_settings["P1i_P2o"] = ConnectionSetting.DIAGONAL
        ws.carrier.connection_settings["P2i_P1o"] = ConnectionSetting.DIAGONAL

        if sim.signal:
            signal_refill = sim.signal.any_frequencies_changing and (
                # only non-zero spaces which aren't changing need to be added
                # if signal frequency is changing
                float(self.L.value) > 0
                or self.L.is_changing
            )
            signal_refill |= self in sim.trace_forest
            signal_refill |= self.user_gouy_x.is_changing
            signal_refill |= self.user_gouy_y.is_changing
            ws.signal.add_fill_function(space_signal_fill, signal_refill)

            ws.signal.connection_settings["P1i_P2o"] = ConnectionSetting.DIAGONAL
            ws.signal.connection_settings["P2i_P1o"] = ConnectionSetting.DIAGONAL
            ws.signal.connection_settings["SIGPHS_P1o"] = ConnectionSetting.DIAGONAL
            ws.signal.connection_settings["SIGPHS_P2o"] = ConnectionSetting.DIAGONAL
            ws.signal.connection_settings["SIGAMP_P1o"] = ConnectionSetting.DIAGONAL
            ws.signal.connection_settings["SIGAMP_P2o"] = ConnectionSetting.DIAGONAL
            ws.signal.connection_settings["H_P1o"] = ConnectionSetting.DIAGONAL
            ws.signal.connection_settings["H_P2o"] = ConnectionSetting.DIAGONAL

        # Initialise the ABCD matrix memory-views
        if sim.is_modal:
            ws.abcd = self.ABCD(self.p1.i, self.p2.o, "x", copy=False)

        return ws

    def _check_L(self, value):
        if value < 0:
            raise ValueError("Length of a space must not be negative.")

        return value

    def _check_nr(self, value):
        if value < 1:
            raise ValueError("Index of refraction must be >= 1")

        return value

    def _resymbolise_ABCDs(self):
        L = self.L.ref
        nr = self.nr.ref
        M_sym = abcd.space(L, nr)

        # Matrices same in both propagations and both planes so
        # only need one register call for all couplings here
        self.register_abcd_matrix(
            M_sym,
            (self.p1.i, self.p2.o),
            (self.p2.i, self.p1.o),
        )

    @property
    def abcd(self):
        """Numeric ABCD matrix.

        Equivalent to any of ``space.ABCD(1, 2, "x")``, ``space.ABCD(2, 1, "x")``,
        ``space.ABCD(1, 2, "y")``, ``space.ABCD(2, 1, "y")``.

        :`getter`: Returns a copy of the (numeric) ABCD matrix (read-only).
        """
        return self.ABCD(1, 2, "x")

    def ABCD(
        self,
        from_node,
        to_node,
        direction="x",
        symbolic=False,
        copy=True,
        retboth=False,
        allow_reverse=False,
    ):
        r"""Returns the ABCD matrix of the space for the specified coupling.

        .. image:: /images/space/abcd_spacet.*
            :align: center

        This is given by,

        .. math::
            M = \begin{pmatrix}
                    1 & \frac{L}{n_r} \\
                    0 & 1
                \end{pmatrix},

        where :math:`L` is the length of the space and :math:`n_r` is
        the index of refraction.

        See :meth:`.Connector.ABCD` for descriptions of parameters, return values and possible
        exceptions.
        """
        return super().ABCD(
            from_node, to_node, direction, symbolic, copy, retboth, allow_reverse
        )

    # replacing spaces makes no sense, you can only replace with another space
    def replace(self, *args, **kwargs) -> str:
        raise NotImplementedError(
            "Replacing a space component is currently not implemented. Consider "
            "breaking up the space into two spaces and a Nothing component to insert "
            "new components in the middle of a space"
        )



#cython: boundscheck=False, wraparound=False, initializedcheck=False

cimport numpy as np
import numpy as np

import logging

from ...cymath cimport complex_t
from ...cymath.complex cimport conj, cexp, carg
from ...cymath.complex cimport cnorm, COMPLEX_0
from ...cymath.math cimport sqrt
from ...knm cimport KnmMatrix
from ...components.node import NodeType
from ...simulations.homsolver cimport HOMSolver
from ...env import warn
from ..workspace cimport (
    DetectorWorkspace,
    MaskedDetectorWorkspace,
    OutputFuncWrapper,
)


LOGGER = logging.getLogger(__name__)

### Amplitude detector workspace & output funcs ###

cdef class ADWorkspace(MaskedDetectorWorkspace):
    cdef public:
        object ntype
        double scaling
        bint is_f_changing
        Py_ssize_t node_id
        Py_ssize_t freq_idx
        Py_ssize_t hom_idx
        HOMSolver solver

    def __init__(self, owner, sim):
        self.is_f_changing = owner.f.is_changing
        if owner.f.eval() is None:
            raise ValueError(f"{owner.f}: frequency value is `None`, check values have been set correctly.")
        fval = float(owner.f)
        fs = []

        for mtx in (sim.carrier, sim.signal):
            if mtx:
                f = mtx.get_frequency_object(fval, owner.node)
                if f is not None:
                    fs.append((f, mtx))
        if len(fs) == 0:
            raise Exception(
                f"Error in amplitude detector {owner.name}:\n"
                f"    Could not find a frequency bin at {owner.f}"
            )
        elif len(fs) > 1:
            raise Exception(
                f"Error in amplitude detector {owner.name}:\n"
                f"    Found multiple frequency bins at {owner.f}"
            )

        freq, self.solver = fs[0]

        if self.solver is sim.carrier:
            super().__init__(owner, sim, needs_carrier=True)
        else:
            super().__init__(owner, sim, needs_signal=True)

        cdef bint multi_field_mode = False
        cdef Py_ssize_t hom_idx = 0
        if owner.node.type == NodeType.OPTICAL:
            if self.sim.is_modal:
                multi_field_mode = owner.n is None and owner.m is None
                if not multi_field_mode:
                    try:
                        hom_idx = self.sim.model.mode_index_map[(owner.n, owner.m)]
                    except KeyError:
                        raise Exception(
                            f"Error in amplitude detector {owner.name}:\n"
                            f"    HOM (n={owner.n}, m={owner.m}) not found in the model"
                        )
        self.freq_idx = freq.index
        self.node_id = self.solver.node_id(owner.node)
        self.hom_idx = hom_idx

        if not multi_field_mode:
            if self.has_mask and not self.hom_in_modes(hom_idx):
                self.set_output_fn(ad_single_field_output_masked)
                warn(
                    f"Masking {repr(owner.name)} which is in single-frequency mode. "
                    f"This will always return values of complex zero in a simulation."
                )
            else:
                if self.has_mask:
                    warn(f"Mask applied to {repr(owner.name)} has no effect!")

                if freq.audio_order == -1:
                    self.set_output_fn(ad_conj_single_field_output)
                else:
                    self.set_output_fn(ad_single_field_output)
        else:
            if freq.audio_order == -1:
                if self.has_mask:
                    self.set_output_fn(ad_conj_multi_field_output_masked)
                else:
                    self.set_output_fn(ad_conj_multi_field_output)
            else:
                if self.has_mask:
                    self.set_output_fn(ad_multi_field_output_masked)
                else:
                    self.set_output_fn(ad_multi_field_output)

        if owner.node.type == NodeType.OPTICAL:
            self.scaling = sqrt(0.5 * self.sim.model_settings.EPSILON0_C)
        elif owner.node.type == NodeType.MECHANICAL:
            self.scaling = self.sim.model_settings.x_scale
        else:
            self.scaling = 1


ad_single_field_output = OutputFuncWrapper.make_from_ptr(c_ad_single_field_output)
cdef c_ad_single_field_output(DetectorWorkspace dws) :
    cdef ADWorkspace ws = <ADWorkspace> dws
    cdef Py_ssize_t N = 0
    cdef complex_t *ptr = ws.solver.node_field_vector_fast(ws.node_id, ws.freq_idx, &N)
    assert(N > 0)
    assert(ptr != NULL)
    return ws.scaling * ptr[ws.hom_idx]

ad_conj_single_field_output = OutputFuncWrapper.make_from_ptr(c_ad_conj_single_field_output)
cdef c_ad_conj_single_field_output(DetectorWorkspace dws) :
    cdef ADWorkspace ws = <ADWorkspace> dws
    cdef Py_ssize_t N = 0
    cdef complex_t *ptr = ws.solver.node_field_vector_fast(ws.node_id, ws.freq_idx, &N)
    assert(N > 0)
    assert(ptr != NULL)
    return ws.scaling * conj(ptr[ws.hom_idx])

ad_single_field_output_masked = OutputFuncWrapper.make_from_ptr(c_ad_single_field_output_masked)
cdef c_ad_single_field_output_masked(DetectorWorkspace dws) :
    return COMPLEX_0


ad_multi_field_output = OutputFuncWrapper.make_from_ptr(c_ad_multi_field_output)
cdef c_ad_multi_field_output(DetectorWorkspace dws) :
    cdef:
        ADWorkspace ws = <ADWorkspace> dws
        complex_t fields_sum = COMPLEX_0
        double amp = 0.0
        double phase = 0.0
        Py_ssize_t i # field index
        Py_ssize_t N = 0
        complex_t *ptr = NULL

    ptr = ws.solver.node_field_vector_fast(ws.node_id, ws.freq_idx, &N)
    assert(N > 0)
    assert(ptr != NULL)

    for i in range(N):
        fields_sum += ptr[i]
        amp += cnorm(ptr[i])

    phase = carg(fields_sum)
    amp = sqrt(amp)
    return ws.scaling * amp * cexp(1j * phase)

ad_multi_field_output_masked = OutputFuncWrapper.make_from_ptr(c_ad_multi_field_output_masked)
cdef c_ad_multi_field_output_masked(DetectorWorkspace dws) :
    cdef:
        ADWorkspace ws = <ADWorkspace> dws
        complex_t field_amp = COMPLEX_0
        complex_t fields_sum = COMPLEX_0
        double amp = 0.0
        double phase = 0.0
        Py_ssize_t i # field index
        Py_ssize_t N = 0
        complex_t *ptr = ws.solver.node_field_vector_fast(ws.node_id, ws.freq_idx, &N)

    assert(N > 0)
    assert(ptr != NULL)

    for i in range(ws.num_unmasked_HOMs):
        field_amp = ptr[ws.unmasked_mode_indices[i]]
        fields_sum += field_amp
        amp += cnorm(field_amp)

    phase = carg(fields_sum)
    amp = sqrt(amp)

    return ws.scaling * amp * cexp(1j * phase)

ad_conj_multi_field_output = OutputFuncWrapper.make_from_ptr(c_ad_conj_multi_field_output)
cdef c_ad_conj_multi_field_output(DetectorWorkspace dws) :
    cdef:
        ADWorkspace ws = <ADWorkspace> dws
        complex_t field_amp = COMPLEX_0
        complex_t fields_sum = COMPLEX_0
        double amp = 0.0
        double phase = 0.0
        Py_ssize_t i # field index
        Py_ssize_t N = 0
        complex_t *ptr = ws.solver.node_field_vector_fast(ws.node_id, ws.freq_idx, &N)

    assert(N > 0)
    assert(ptr != NULL)

    for i in range(ws.sim.model_settings.num_HOMs):
        field_amp = conj(ptr[i])
        fields_sum += field_amp
        amp += cnorm(field_amp)

    phase = carg(fields_sum)
    amp = sqrt(amp)

    return ws.scaling * amp * cexp(1j * phase)

ad_conj_multi_field_output_masked = OutputFuncWrapper.make_from_ptr(c_ad_conj_multi_field_output_masked)
cdef c_ad_conj_multi_field_output_masked(DetectorWorkspace dws) :
    cdef:
        ADWorkspace ws = <ADWorkspace> dws
        complex_t field_amp = COMPLEX_0
        complex_t fields_sum = COMPLEX_0
        double amp = 0.0
        double phase = 0.0
        Py_ssize_t i # field index
        Py_ssize_t N = 0
        complex_t *ptr = ws.solver.node_field_vector_fast(ws.node_id, ws.freq_idx, &N)

    assert(N > 0)
    assert(ptr != NULL)

    for i in range(ws.num_unmasked_HOMs):
        field_amp = conj(ptr[ws.unmasked_mode_indices[i]])
        fields_sum += field_amp
        amp += cnorm(field_amp)

    phase = carg(fields_sum)
    amp = sqrt(amp)

    return ws.scaling * amp * cexp(1j * phase)

### Knm detector workspace & output funcs ###

cdef class KnmDetectorWorkspace(DetectorWorkspace):
    cdef public:
        Py_ssize_t from_idx, to_idx

        KnmMatrix knm_matrix

    def __init__(self, owner, sim):
        super().__init__(owner, sim, needs_modal_update=True)

knm_detector_scalar_output = OutputFuncWrapper.make_from_ptr(c_knm_detector_scalar_output)
cdef c_knm_detector_scalar_output(DetectorWorkspace dws) :
    """Outputs single coefficient of (n1, m1) -> (n2, m2)."""
    cdef KnmDetectorWorkspace kdws = <KnmDetectorWorkspace> dws
    return kdws.knm_matrix.data_view[kdws.to_idx][kdws.from_idx]

knm_detector_mode1_output = OutputFuncWrapper.make_from_ptr(c_knm_detector_mode1_output)
cdef c_knm_detector_mode1_output(DetectorWorkspace dws) :
    """Outputs coefficient vector of (n1, m1) -> (n, m) for each n, m."""
    cdef KnmDetectorWorkspace kdws = <KnmDetectorWorkspace> dws
    return np.asarray(kdws.knm_matrix.data_view[:,kdws.from_idx].copy())

knm_detector_mode2_output = OutputFuncWrapper.make_from_ptr(c_knm_detector_mode2_output)
cdef c_knm_detector_mode2_output(DetectorWorkspace dws) :
    """Outputs coefficient vector of (n, m) -> (n2, m2) for each n, m."""
    cdef KnmDetectorWorkspace kdws = <KnmDetectorWorkspace> dws
    return np.asarray(kdws.knm_matrix.data_view[kdws.to_idx].copy())

knm_detector_matrix_output = OutputFuncWrapper.make_from_ptr(c_knm_detector_matrix_output)
cdef c_knm_detector_matrix_output(DetectorWorkspace dws) :
    """Outputs full matrix of coupling coefficients."""
    cdef KnmDetectorWorkspace kdws = <KnmDetectorWorkspace> dws
    return kdws.knm_matrix.data.copy()



"""Kat script generator."""

import logging
from io import StringIO
from functools import singledispatchmethod
from itertools import chain
from collections.abc import Iterable
import inspect

import numpy as np

from ..env import warn
from ..model import Model
from ..element import ModelElement
from ..parameter import Parameter, ParameterRef
from ..symbols import OPERATORS, FUNCTIONS, Function, Constant
from ..enums import ModulatorType
from ..components import Port, Node
from ..utilities.functools import flagdispatchmethod
from ..components.general import LocalDegreeOfFreedom
from ..analysis.actions.base import Action
from .containers import (
    KatCoordinate,
    KatToken,
    KatNumberToken,
    KatStringToken,
    KatWhitespaceToken,
    KatNoneToken,
    KatFixMeToken,
    KatScript,
    KatElement,
    KatFunction,
    KatKwarg,
    KatExpression,
    KatGroupedExpression,
    KatArray,
    FixMeValue,
)
from .tokens import LITERALS
from .adapter import ArgumentDump, ArgumentType, ElementDump, CommandDump, AnalysisDump
from .graph import ROOT_NODE_NAME, KatNodeType, KatEdgeType, KatGraph
from .util import scriptsorted, merge_attributes


LOGGER = logging.getLogger(__name__)


class _empty:
    """Represents an unset value.

    This exists for use when `None` is considered non-default.
    """


def _first_instruction_token(node_data):
    """Get the first occurring token in `node_data`.

    This is usually `node_data["token"]` but for e.g. grouped expressions it is the
    first "(" in `node_data["extra_tokens"]`.
    """
    tokens = []
    if token := node_data.get("token"):
        tokens.append(token)
    if extras := node_data.get("extra_tokens"):
        tokens.extend(extras)
    return next(iter(scriptsorted(tokens)))


class ElementContainer:
    """Container for top level model elements, used by the generator."""

    def __init__(self, element):
        self.element = element

    def __repr__(self):
        return f"<{self.__class__.__name__}({repr(self.element)})>"


class CommandContainer:
    """Container for top level commands, used by the generator."""

    def __init__(self, adapter, params):
        self.adapter = adapter
        self.params = params

    def __repr__(self):
        return f"<{self.__class__.__name__}({self.adapter.full_name})>"


class KatUnbuilder:
    """Model to KatScript converter.

    This class defines single-dispatch methods to recursively fill Finesse objects into
    a :class:`.KatGraph`. The graph is then serialised to KatScript via
    :class:`.KatUnfiller`.

    The single-dispatch uses the type of the object to create appropriate nodes and
    edges in the :class:`.KatGraph`.

    Typically unbuilding starts with the passing of a :class:`.Model` to
    :meth:`~.KatUnbuilder.unbuild` or :meth:`~.KatUnbuilder.unbuild_file`. The
    single-dispatch method that handles models then extracts the elements, commands and
    analyses and passes them recursively to the same single-dispatch, whereupon matching
    fill methods further handle them. The underlying tokens that form Finesse objects
    are eventually reached by recursively calling the fill method, and these are added
    as terminal nodes in the graph and connected to their parents by edges.

    Special behaviour applies in some cases. One such case is the handling of objects
    from containers with corresponding type annotations specifying that they are
    :class:`.Parameter` objects. These are assumed to have *name* KatScript
    representation, rather than value. This is used for example in the dumping of axis
    parameters, which must be written as e.g. `xaxis(m1.phi, ...)` rather than `xaxis(0,
    ...)` (in the case that `m1.phi` has value `0`).
    """

    # Model element types not dumped.
    IGNORED_ELEMENTS = [
        # Models always contain an Fsig. The actual value is dumped as a command.
        "Fsig",
    ]

    # FIXME: move to KatSpec.
    UNARY_OPERATORS = {
        "pos": "+",
        "neg": "-",
    }

    # Reverse dicts for various types.
    LITERAL_MAP = dict((value, key) for key, value in LITERALS.items())

    def __init__(self, show_warnings=True, spec=None):
        if spec is None:
            from .spec import KATSPEC as spec

        self.spec = spec
        self.graph = None
        self._position = None
        self._stack = None  # Tracks _fill level.
        self._has_fixmes = False
        self.show_warnings = show_warnings
        self.__tmp_regen_warnings = set()  # Warnings for not-yet-implemented regens.

    def unbuild(self, *args, **kwargs):
        fobj = StringIO()
        self.unbuild_file(fobj, *args, **kwargs)
        fobj.seek(0)
        return fobj.read()

    def unbuild_file(
        self,
        fobj,
        item,
        ref_graph=None,
        ref_node=None,
        directive_defaults=False,
        argument_defaults=False,
        prefer_keywords=True,
    ):
        """Generate KatScript for `item` and write to `fobj`.

        Parameters
        ----------
        fobj : :class:`io.FileIO`
            The file object to write KatScript to. This should be opened in text mode.

        item : object
            The object to generate KatScript for. This is normally a :class:`.Model` but
            can be any Finesse object.

        ref_graph : :class:`.KatGraph`, optional
            The reference syntax graph to use during KatScript generation. If specified,
            the nodes representing `item` and any Finesse objects contained therein will
            be copied and updated from it. This preserves formatting such as whitespace
            and comments.

        ref_node : :class:`str`, optional
            The reference node to use when unparsing `item`. This is only used if
            `ref_graph` is specified. If `ref_graph` is given but `ref_node` is not,
            `ref_node` is assumed to be the graph root. If the graph does not contain
            a root node, a :class:`ValueError` is raised.

        directive_defaults : :class:`bool`, optional
            Generate default elements, commands and actions even if they are set to
            their default values and would therefore generate Finesse objects in the
            same state when later parsed. If `ref_graph` is specified and an element,
            command or action is present as a node, however, it is generated regardless
            of whether its value is the default or not. Defaults to `False`.

        argument_defaults : :class:`bool`, optional
            Generate optional arguments that are set to their default values and would
            therefore generate Finesse objects in the same state when later parsed. If
            `ref_graph` is specified and a keyword argument is present as a node,
            however, it is generated regardless of whether its value is the default or
            not. Defaults to `False`.

        prefer_keywords : :class:`bool`, optional
            Prefer generating keyword arguments over positional arguments, where
            allowed. If a parameter can be a keyword argument, i.e. it is not
            positional-only, it is dumped as a keyword argument if this is `True`. This
            setting is ignored if `ref_node` is specified and a reference argument for
            the parameter is found (whereupon the original style is reused). Defaults to
            `True`.
        """
        self._stack = []

        if ref_graph is not None:
            if ref_node is None:
                # Let the user specify the starting node if they know what they're
                # doing.
                if ROOT_NODE_NAME not in ref_graph:
                    raise ValueError(
                        f"if ref_node is not specified, ref_graph must contain a root "
                        f"node named {ROOT_NODE_NAME}"
                    )
                ref_node = ROOT_NODE_NAME

            self._debug(
                f"using {repr(ref_graph)} at node {repr(ref_node)} as a reference to "
                f"unbuild {repr(item)}"
            )

        self.graph = KatGraph()
        self.ref_graph = ref_graph
        self.directive_defaults = directive_defaults
        self.argument_defaults = argument_defaults
        self.prefer_keywords = prefer_keywords
        self.position = KatCoordinate(1, 1)
        self._has_fixmes = False
        self.__tmp_regen_warnings.clear()

        # Fill the graph.
        if self._fill(item, ROOT_NODE_NAME, ref_node=ref_node):
            # Generate KatScript from the graph.
            unfiller = KatUnfiller()
            unfiller.unfill_file(fobj, ROOT_NODE_NAME, self.graph)
        else:
            self._debug(f"no script generated for {repr(item)}")

        # Trigger log messages for unimplemented intent-preserving unparsing behaviours.
        # This should be removed eventually once regeneration works for all KatScript
        # patterns.
        if self.show_warnings:
            for msg in self.__tmp_regen_warnings:
                # User can't really do anything about this, so trigger a log message
                # instead of a proper warning.
                LOGGER.warning(msg)

        # Trigger a warning if a "fixme" value was dumped.
        if self._has_fixmes:
            warn(
                "The generated script contains invalid values. Search the script for "
                "__FIXME__ placeholders and replace appropriately."
            )

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, position):
        self._debug(f"setting position to {position}")
        self._position = position

    @property
    def lineno(self):
        return self.position.lineno

    @property
    def index(self):
        return self.position.index

    @property
    def _stackpos(self):
        return ".".join([t.__name__ for t in self._stack])

    def _debug(self, msg):
        if not LOGGER.isEnabledFor(logging.DEBUG):
            return
        caller = inspect.getouterframes(inspect.currentframe(), 2)[1][3]  # Sloooow.
        LOGGER.debug(f"[{self._stackpos}::{caller}] {msg}", stacklevel=2)

    def _fill(self, value, path, ref_node=None, **attributes):
        """Fill `value` into the graph at `path`, optionally reusing unchanged tokens
        from an existing graph node."""
        assert path not in self.graph
        self._stack.append(type(value))
        self._debug(f"filling {repr(value)} into graph at {repr(path)}")

        ref_extras = None
        if ref_node is not None:
            #  Get the reference node's extras.
            if node_extras := self.ref_graph.nodes[ref_node].get("extra_tokens"):
                # This object gets mutated by filling methods, so we therefore cast
                # it to a list so we don't run into issues with graph views,
                # generators, etc.
                ref_extras = list(scriptsorted(node_extras))

        if filled_attributes := self._do_fill(
            value, path, ref_node=ref_node, ref_extras=ref_extras
        ):
            attributes = merge_attributes(attributes, filled_attributes)
            self.graph.add_node(path, **attributes)
        else:
            self._debug(f"nothing to add for {value}")

        self._stack.pop()

        return attributes

    @singledispatchmethod
    def _do_fill(self, item, path, ref_node, ref_extras):
        """Get the attributes for `item` to be added to the corresponding node."""
        # This is the default dispatch when the type isn't recognised.
        raise NotImplementedError(
            f"don't know how to generate KatScript for {repr(item.__class__)}"
        )

    @_do_fill.register(Model)
    def _(self, model, path, ref_node, ref_extras):
        extra_tokens = []
        order = 0

        # Ask each registered adapter for its dumps from this model.
        element_dumps = {}
        for adapter in set(self.spec.elements.values()):
            if adapter.getter is None:
                continue

            self._debug(f"dumping items for {adapter.full_name}")
            for dump in adapter.getter(adapter, model):
                element_dumps[dump.element.name] = dump

        singular_function_dumps = {}
        nonsingular_function_dumps = []
        functions = set(self.spec.analyses.values()) | set(self.spec.commands.values())
        for adapter in functions:
            if adapter.getter is None:
                continue

            self._debug(f"dumping items for {adapter.full_name}")
            for dump in adapter.getter(adapter, model):
                if adapter.singular:
                    singular_function_dumps[adapter.full_name] = dump
                else:
                    nonsingular_function_dumps.append(dump)

        if ref_node is not None:
            for node, data in self.ref_graph.sorted_dependent_argument_nodes(
                ref_node, data=True
            ):
                if data["type"] is KatNodeType.ELEMENT:
                    element_name = data["name_token"].raw_value

                    try:
                        dump = element_dumps.pop(element_name)
                    except KeyError:
                        self._debug(
                            f"reference element {repr(element_name)} no longer present"
                        )
                        continue
                    else:
                        self._debug(
                            f"reference element {repr(element_name)} still present"
                        )
                elif data["type"] is KatNodeType.FUNCTION:
                    # Commands and analyses.
                    function_name = data["token"].raw_value
                    adapter = self.spec.function_directives[function_name]

                    if adapter.singular:
                        try:
                            dump = singular_function_dumps.pop(adapter.full_name)
                        except KeyError:
                            self._debug(
                                f"reference function {repr(function_name)} no longer "
                                f"present"
                            )
                            continue
                        else:
                            self._debug(
                                f"reference function {repr(function_name)} still "
                                f"present"
                            )
                    else:
                        self.__tmp_regen_warnings.add(
                            "Intent-preserving regeneration of KatScript for "
                            "nonsingular commands has not been implemented yet. These "
                            "have instead been generated using default formatting "
                            "(this does not affect the validity of the generated "
                            "KatScript)."
                        )
                        continue
                else:
                    raise ValueError("unexpected dependent argument node type")

                # Reuse any extra tokens specified in the argument's parent up until
                # `token`, ensuring there is a delimiter.
                extra_tokens.extend(
                    self._handle_script_arg_extras(order, data["token"], ref_extras)
                )

                argument_path = self.graph.item_node_name(order, path)
                if self._fill(dump, argument_path, node):
                    self.graph.add_edge(
                        argument_path, path, type=KatEdgeType.ARGUMENT, order=order
                    )

                    order += 1

        # Add any remaining extras.
        extra_tokens.extend(self._maybe_reuse_extras(ref_extras))

        ###
        # Now deal with stuff that was added since the last parse.
        ###

        if order > 0 and (
            element_dumps or singular_function_dumps or nonsingular_function_dumps
        ):
            # There were reference items dumped first. Add a comment to say that the
            # next items are new.
            extra_tokens.append(self._use_token(self._newline_token()))
            extra_tokens.append(
                self._use_token(
                    self._comment_token(
                        "Items below could not be matched to original script, or were "
                        "not present when the model was originally parsed."
                    )
                )
            )

        # Components, detectors, etc.
        # Dump the elements in the order in which they were added to the model.
        sorted_element_dumps = sorted(
            element_dumps.values(),
            # if item isn't in the element dict (a model parameter) then use -1?
            key=lambda dump: (
                model.element_order(dump.element)
                if dump.element in model.elements
                else -1
            ),
        )
        for element_dump in sorted_element_dumps:
            if order > 0:
                extra_tokens.append(self._use_token(self._newline_token()))

            element_path = self.graph.item_node_name(order, path)
            if self._fill(element_dump, element_path):
                self.graph.add_edge(
                    element_path, path, type=KatEdgeType.ARGUMENT, order=order
                )

                order += 1

        # Function dumps.
        for dump in chain(singular_function_dumps.values(), nonsingular_function_dumps):
            if order > 0:
                extra_tokens.append(self._use_token(self._newline_token()))

            function_path = self.graph.item_node_name(order, path)
            if self._fill(dump, function_path):
                self.graph.add_edge(
                    function_path, path, type=KatEdgeType.ARGUMENT, order=order
                )

                order += 1

        return {"type": KatNodeType.ROOT, "extra_tokens": extra_tokens}

    @_do_fill.register(ElementDump)
    def _(self, dump, path, ref_node, ref_extras):
        """Model element definitions."""
        if dump.element.__class__.__name__ in self.IGNORED_ELEMENTS:
            self._debug(f"skipping filling of ignored element {dump.element}")
            return

        # Ignore default values unless they were explicitly defined in the reference.
        if not self.directive_defaults and dump.is_default:
            self._debug(
                f"ignoring {repr(dump.item_name)} as its parameters are set to their "
                f"default values (set `directive_defaults` to change this behaviour)"
            )
            return

        extra_tokens = []

        if ref_node is not None:
            data = self.ref_graph.nodes[ref_node]
            ref_token = data["token"]
            ref_name_token = data["name_token"]

            token = self._reuse_token(ref_token)

            # Fill in extras between directive and name.
            extra_tokens.extend(
                self._maybe_reuse_extras(
                    ref_extras,
                    until_token=ref_name_token,
                )
            )

            name_token = self._reuse_token(ref_name_token)

            # Fill in the arguments.
            extra_tokens.extend(
                self._fill_element_args(
                    dump.parameters, path, ref_node=ref_node, ref_extras=ref_extras
                )
            )

            # Add any remaining extras.
            extra_tokens.extend(self._maybe_reuse_extras(ref_extras))
        else:
            token = self._use_token(self._name_token(dump.adapter.full_name))
            extra_tokens.append(self._use_token(self._space_token()))
            name_token = self._use_token(self._name_token(dump.element.name))
            extra_tokens.extend(self._fill_element_args(dump.parameters, path))

        return {
            "token": token,
            "name_token": name_token,
            "type": KatNodeType.ELEMENT,
            "extra_tokens": extra_tokens,
        }

    @_do_fill.register(AnalysisDump)
    @_do_fill.register(CommandDump)
    def _(self, dump, path, ref_node, ref_extras):
        # Ignore default values unless they were explicitly defined in the reference.
        if not self.directive_defaults and dump.is_default:
            self._debug(
                f"ignoring {repr(dump.item_name)} as its parameters are set to their "
                f"default values (set `directive_defaults` to change this behaviour)"
            )
            return

        if ref_node is not None:
            token = self._reuse_token(self.ref_graph.nodes[ref_node]["token"])
        else:
            token = self._use_token(self._name_token(dump.adapter.full_name))

        extra_tokens = self._fill_group(
            dump.parameters,
            path,
            "()",
            self._fill_function_args,
            ref_node=ref_node,
            ref_extras=ref_extras,
        )

        return {
            "token": token,
            "type": KatNodeType.FUNCTION,
            "extra_tokens": extra_tokens,
        }

    @_do_fill.register(Action)
    def _(self, action, path, ref_node, ref_extras):
        """A nested action.

        This is NOT for top level actions, which must be dumped as an
        :class:`.AnalysisDump` object.
        """
        # Ensure this isn't a top level action.
        assert path.count(".") > 1, "disallowed top level action"

        # Find the adapter willing to generate a dump for this action.
        for adapter in self.spec.analyses.values():
            if adapter.getter is None:
                continue

            try:
                return self._fill(
                    next(adapter.getter.dump(adapter, action)), path, ref_node
                )
            except StopIteration:
                pass

        raise NotImplementedError(
            f"don't know how to generate KatScript for {repr(action.__class__)}"
        )

    @_do_fill.register(ArgumentDump)
    def _(self, argument, path, ref_node, ref_extras):
        if argument.reference:
            # This represents a target that should be dumped by name.
            return {
                "token": self._maybe_reuse_value(
                    argument.value.full_name, self._name_token, ref_node=ref_node
                ),
                "type": KatNodeType.VALUE,
            }

        return self._fill(argument.value, path, ref_node)

    @_do_fill.register(Parameter)
    def _(self, parameter, path, ref_node, ref_extras):
        if parameter.is_externally_controlled:
            return self._fill(parameter._get_unset_value(), path, ref_node)
        else:
            return self._fill(parameter.value, path, ref_node)

    @_do_fill.register(ParameterRef)
    def _(self, reference, path, ref_node, ref_extras):
        if reference.parameter.is_default_for_owner:
            # We can generate without the ".value" part.
            value = reference.owner.name
        else:
            value = reference.name

        return {
            "token": self._maybe_reuse_value(
                value, self._name_token, ref_node=ref_node
            ),
            "type": KatNodeType.REFERENCE,
        }

    @_do_fill.register(ModelElement)
    def _(self, value, path, ref_node, ref_extras):
        """Reference to a model element.

        Model element definitions are matched as :class:`.ElementContainer`.
        """
        return {
            "token": self._maybe_reuse_value(
                value.name, self._name_token, ref_node=ref_node
            ),
            "type": KatNodeType.VALUE,
        }

    @_do_fill.register(Port)
    @_do_fill.register(Node)
    def _(self, value, path, ref_node, ref_extras):
        return {
            "token": self._maybe_reuse_value(
                value.full_name, self._name_token, ref_node=ref_node
            ),
            "type": KatNodeType.VALUE,
        }

    @_do_fill.register(LocalDegreeOfFreedom)
    def _(self, definition, path, ref_node, ref_extras):
        return {
            "token": self._maybe_reuse_value(
                definition.name, self._name_token, ref_node=ref_node
            ),
            "type": KatNodeType.VALUE,
        }

    @_do_fill.register(ModulatorType)
    def _(self, mod_type, path, ref_node, ref_extras):
        return {
            "token": self._maybe_reuse_value(
                mod_type.name, self._name_token, ref_node=ref_node
            ),
            "type": KatNodeType.VALUE,
        }

    @_do_fill.register(int)
    @_do_fill.register(np.integer)
    def _(self, value, path, ref_node, ref_extras):
        value = int(value)
        if value < 0:
            return self._fill(FUNCTIONS["neg"](abs(value)), path, ref_node)

        return {
            "token": self._maybe_reuse_value(
                value, self._number_token, ref_node=ref_node
            ),
            "type": KatNodeType.VALUE,
        }

    @_do_fill.register(float)
    @_do_fill.register(np.floating)
    def _(self, value, path, ref_node, ref_extras):
        # Kinda strange, but isinstance(np.nan, float) == True, so we need to deal with
        # that.
        if np.isnan(value):
            # Dump a "fixme" value instead.
            return self._fill(FixMeValue, path, ref_node)

        value = float(value)
        if value < 0:
            return self._fill(FUNCTIONS["neg"](abs(value)), path, ref_node)

        return {
            "token": self._maybe_reuse_value(
                value, self._number_token, ref_node=ref_node
            ),
            "type": KatNodeType.VALUE,
        }

    @_do_fill.register(complex)
    def _(self, value, path, ref_node, ref_extras):
        real = value.real
        imag = value.imag

        if not real:
            cplx = complex(f"{abs(imag)}j")
            if imag < 0:
                return self._fill(FUNCTIONS["neg"](cplx), path, ref_node)
            return {
                "token": self._maybe_reuse_value(
                    cplx, self._number_token, ref_node=ref_node
                ),
                "type": KatNodeType.VALUE,
            }
        elif not imag:
            if real < 0:
                return self._fill(FUNCTIONS["neg"](abs(real)), path, ref_node)
            return {
                "token": self._maybe_reuse_value(
                    real, self._number_token, ref_node=ref_node
                ),
                "type": KatNodeType.VALUE,
            }
        else:
            if real < 0:
                real = FUNCTIONS["neg"](abs(real))
            binop = OPERATORS["__add__"] if imag >= 0 else OPERATORS["__sub__"]
            return self._fill(binop(real, complex(f"{abs(imag)}j")), path, ref_node)

    @_do_fill.register(str)
    def _(self, value, path, ref_node, ref_extras):
        if value in self.spec.keywords:
            factory = self._name_token
        else:
            factory = self._string_token

        return {
            "token": self._maybe_reuse_value(value, factory, ref_node=ref_node),
            "type": KatNodeType.VALUE,
        }

    @_do_fill.register(bool)
    @_do_fill.register(np.bool_)
    def _(self, value, path, ref_node, ref_extras):
        return {
            "token": self._maybe_reuse_value(
                "true" if value else "false", self._name_token, ref_node=ref_node
            ),
            "type": KatNodeType.VALUE,
        }

    @_do_fill.register(type(None))
    def _(self, value, path, ref_node, ref_extras):
        return {
            "token": self._maybe_reuse_value(None, self._none_token, ref_node=ref_node),
            "type": KatNodeType.VALUE,
        }

    @_do_fill.register(type(FixMeValue))
    def _(self, value, path, ref_node, ref_extras):
        self._has_fixmes = True

        return {
            "token": self._maybe_reuse_value(
                FixMeValue, self._fixme_token, ref_node=ref_node
            ),
            "type": KatNodeType.VALUE,
        }

    @_do_fill.register(Iterable)
    def _(self, array, path, ref_node, ref_extras):
        extra_tokens = self._fill_group(
            array,
            path,
            "[]",
            self._fill_inline_function_args,
            ref_node=ref_node,
            ref_extras=ref_extras,
        )
        return {"type": KatNodeType.ARRAY, "extra_tokens": extra_tokens}

    @_do_fill.register(Constant)
    def _(self, constant, path, ref_node, ref_extras):
        return self._fill(constant.eval(), path, ref_node)

    @_do_fill.register(Function)
    def _(self, operation, path, ref_node, ref_extras):
        op = operation.name
        extra_tokens = []

        def make_arg(argument, order):
            argument_path = self.graph.item_node_name(order, path)

            # FIXME: add ref_node (should be the argument order child of ref_node, but
            # this is not guaranteed to exist so needs special handling)
            self._fill(argument, argument_path)
            self.graph.add_edge(
                argument_path, path, type=KatEdgeType.ARGUMENT, order=order
            )

        if ref_node is not None:
            self.__tmp_regen_warnings.add(
                "Intent-preserving regeneration of KatScript for operations has not "
                "been implemented yet. These have instead been generated using default "
                "formatting (this does not affect the validity of the generated "
                "KatScript)."
            )

        # In certain cases we get a '-' instead of 'neg' which can be both a
        # unary and binary operator, so check explicitly that args is at least 2
        if op in self.spec.binary_operators and len(operation.args) >= 2:
            if len(operation.args) > 2:
                LOGGER.debug(f"Converting n-ary expression {repr(operation)} to binary")
                operation = operation.to_binary_add_mul()
            assert len(operation.args) == 2
            extra_tokens.append(self._use_token(self._literal_token("(")))
            make_arg(operation.args[0], 0)
            token = self._use_token(self._literal_token(op))
            nodetype = KatNodeType.EXPRESSION
            make_arg(operation.args[1], 1)
            extra_tokens.append(self._use_token(self._literal_token(")")))
        elif op in self.spec.unary_operators:  # op is '+' or '-'
            token = self._use_token(self._literal_token(op))
            nodetype = KatNodeType.FUNCTION
            make_arg(operation.args[0], 0)
        elif unary_op := self.UNARY_OPERATORS.get(op):  # op is 'pos' or 'neg'
            token = self._use_token(self._literal_token(unary_op))
            nodetype = KatNodeType.FUNCTION
            make_arg(operation.args[0], 0)
        elif op in self.spec.expression_functions:
            token = self._use_token(self._name_token(op))
            nodetype = KatNodeType.FUNCTION
            extra_tokens.append(self._use_token(self._literal_token("(")))
            for order, argument in enumerate(operation.args):
                if order > 0:
                    extra_tokens.append(self._use_token(self._literal_token(",")))
                    extra_tokens.append(self._use_token(self._space_token()))
                make_arg(argument, order)
            extra_tokens.append(self._use_token(self._literal_token(")")))
        else:
            raise NotImplementedError(
                f"don't know how to generate KatScript for {repr(op)}"
            )

        return {"token": token, "type": nodetype, "extra_tokens": extra_tokens}

    def _maybe_reuse_value(self, ensure_value, value_factory, ref_node=None):
        """Reuse reference value at `ref_node` if it matches `ensure_value`, otherwise
        create a new token using `value_factory`.

        Returns
        -------
        :class:`.KatToken`
            The token.
        """
        if self.ref_graph and ref_node:
            data = self.ref_graph.nodes[ref_node]

            # The originally parsed string.
            ref_value = data["token"].raw_value

            if ensure_value == data["token"].value:  # NOT raw value!
                self._debug(f"value ({repr(ref_value)}) unchanged")
                return self._reuse_token(data["token"])
            else:
                self._debug(
                    f"value changed from {repr(ref_value)} to {repr(ensure_value)})"
                )

        return self._use_token(value_factory(ensure_value))

    def _maybe_reuse_extras(self, extras, until_token=None, ensure_literal=_empty):
        """Reuse tokens in `extras` in order until `until_token` token position is
        encountered, optionally ensuring `ensure_literal` is present (creating a new
        token using :meth:`._literal_token` if not).

        The `extras` object passed in to this method, if not None, is mutated.

        Yields
        ------
        :class:`.KatToken`
            Reused extra token.
        """
        if not self.ref_graph or not extras:
            extras = []

        seen_value = False

        while extras:
            if until_token and extras[0].start >= until_token.start:
                break

            token = extras.pop(0)

            if ensure_literal == token.value:  # NOT raw value!
                seen_value = True

            yield self._reuse_token(token)

        if ensure_literal is not _empty and not seen_value:
            # We've not seen `ensure_literal` yet, so create one.
            yield self._use_token(self._literal_token(ensure_literal))

    def _use_token(self, token):
        """Put a copy of `token` at the current position, updating the current position
        to the end of the token."""
        out = token.to_new_position(self.position)
        self._debug(f"adding token {out} @ {out.bounds}")

        if nlines := token.raw_value.count("\n"):
            self.position = KatCoordinate(self.lineno + nlines, 1)
        else:
            self.position = out.stop

        return out

    def _reuse_token(self, token):
        self._debug(f"reusing reference token {token} originally @ {token.bounds}")
        return self._use_token(token)

    def _name_token(self, value):
        return KatToken(0, 0, len(value), "NAME", value)

    def _space_token(self, length=1):
        return KatWhitespaceToken(0, 0, length, "WHITESPACE", " " * length)

    def _number_token(self, value):
        value = str(value)
        return KatNumberToken(0, 0, len(value), "NUMBER", value)

    def _string_token(self, value):
        value = repr(value)
        return KatStringToken(0, 0, len(value), "STRING", value)

    def _none_token(self, value):
        assert value is None
        value = "none"
        return KatNoneToken(0, 0, len(value), "NONE", value)

    def _fixme_token(self, value):
        assert value == FixMeValue
        value = "__FIXME__"
        return KatFixMeToken(0, 0, len(value), "FIXME", value)

    def _literal_token(self, literal):
        return KatToken(0, 0, len(literal), self.LITERAL_MAP[literal], literal)

    def _newline_token(self, count=1):
        value = "\n" * count
        return KatToken(0, 0, len(value), "NEWLINE", value)

    def _comment_token(self, comment):
        value = f"# {comment}"
        return KatToken(0, 0, len(value), "COMMENT", value)

    def _fill_element_args(self, parameters, path, **kwargs):
        return self._fill_directive_args(
            parameters, path, self._delimit_element_arg, **kwargs
        )

    def _fill_function_args(self, parameters, path, **kwargs):
        return self._fill_directive_args(
            parameters, path, self._delimit_function_arg, **kwargs
        )

    def _fill_directive_args(
        self, parameters, path, delimiter_handler, ref_node=None, **kwargs
    ):
        """Fill a directive's arguments."""
        ref_args, ref_kwargs = self._directive_ref_args(parameters, ref_node)
        arg_map, var_arg, kwarg_map, var_kwarg = self._deal_dump_parameters(
            parameters,
            prefer_keywords=self.prefer_keywords,
            ref_args=ref_args,
            ref_kwargs=ref_kwargs,
        )

        return self._fill_delimited_args(
            arg_map,
            var_arg,
            kwarg_map,
            var_kwarg,
            path,
            ref_args=ref_args,
            ref_kwargs=ref_kwargs,
            delimiter_handler=delimiter_handler,
            **kwargs,
        )

    def _fill_inline_function_args(self, items, path, ref_node=None, **kwargs):
        """Fill inline function arguments, e.g. those of an array or expression
        function.

        Inline functions always use positional arguments.
        """
        ref_args = self._inline_ref_args(items, ref_node)

        # The function filler expects a dict. Give each item a key that can be used to
        # determine whether a reference item exists.
        items = {index: item for index, item in enumerate(items)}

        # Expression functions always use positional arguments.
        return self._fill_delimited_args(
            items,
            None,
            {},
            None,
            path,
            delimiter_handler=self._delimit_function_arg,
            ref_args=ref_args,
            **kwargs,
        )

    def _fill_group(
        self, arguments, path, delimiters, argument_filler, ref_node, ref_extras
    ):
        """Fill a literal-delimited group containing arguments.

        "Literal-delimited" means that the group's outer delimiters are literal
        characters like "(" and ")" or "[" and "]".
        """
        opening_delimiter, closing_delimiter = delimiters
        assert len(opening_delimiter) == len(closing_delimiter) == 1

        def opening_delimiter_from_extras():
            """Grab the opening delimiter for this group from the reference extras, if
            found."""
            nonlocal ref_extras

            if not ref_extras:
                return

            # Assume the opening delimiter is the first extra.
            first_extra = ref_extras.pop(0)

            assert first_extra.raw_value == opening_delimiter
            return first_extra

        extra_tokens = []

        if first_extra := opening_delimiter_from_extras():
            extra_tokens.append(self._reuse_token(first_extra))
        else:
            extra_tokens.append(self._use_token(self._literal_token(opening_delimiter)))

        # Fill in the arguments.
        extra_tokens.extend(
            argument_filler(arguments, path, ref_node=ref_node, ref_extras=ref_extras)
        )

        # Fill in any remaining extras, ensuring there's a closing delimiter.
        extra_tokens.extend(
            self._maybe_reuse_extras(ref_extras, ensure_literal=closing_delimiter)
        )

        return extra_tokens

    def _directive_ref_args(self, parameters, ref_node):
        ref_args = {}
        ref_kwargs = {}

        # Check if any of the originally parsed arguments are still present.
        if ref_node is not None:
            ref_arguments = self.ref_graph.sorted_dependent_argument_nodes(
                ref_node, data=True
            )

            directive_name = self.ref_graph.nodes[ref_node]["token"].raw_value
            adapter = self.spec.directives[directive_name]

            for position, (node, data) in enumerate(ref_arguments, start=1):
                if key_token := data.get("key_token"):
                    # This is a keyword argument.
                    name = key_token.raw_value
                    if name in parameters:
                        self._debug(
                            f"reference keyword argument {repr(name)} still present"
                        )
                        ref_kwargs[name] = node, data
                else:
                    # This is a positional argument. Figure out the name.
                    argument = self.ref_graph.argument(node, adapter)
                    if argument.name in parameters:
                        self._debug(
                            f"reference positional argument {repr(argument.name)} "
                            f"(position {position}) still present"
                        )
                        ref_args[argument.name] = node, data

        return ref_args, ref_kwargs

    def _inline_ref_args(self, items, ref_node):
        ref_args = {}

        # Check if any of the originally parsed arguments are still present.
        if ref_node is not None:
            ref_arguments = self.ref_graph.sorted_dependent_argument_nodes(
                ref_node, data=True
            )

            for position, (node, data) in enumerate(ref_arguments):
                ref_args[position] = node, data

        return ref_args

    def _deal_dump_parameters(
        self, parameters, prefer_keywords, ref_args=None, ref_kwargs=None
    ):
        """Split mapping of parameters to :class:`.ArgumentDump` in `parameters` into
        positional and keyword argument mappings depending on `prefer_keywords`."""
        positional = {}
        keyword = {}

        if ref_args is None:
            ref_args = {}
        if ref_kwargs is None:
            ref_kwargs = {}

        var_pos = None
        var_keyword = None

        # Check if there is a variadic positional argument. If there is, then any
        has_var_pos = any(
            p.kind is ArgumentType.VAR_POS and p.value for p in parameters.values()
        )

        for arg, dump_param in parameters.items():
            if dump_param.kind is ArgumentType.POS_ONLY:
                positional[arg] = dump_param
            elif dump_param.kind is ArgumentType.KEYWORD_ONLY:
                # Ignore default values unless they were explicitly defined in the
                # reference.
                if (
                    not self.argument_defaults
                    and arg not in ref_kwargs
                    and dump_param.is_default
                ):
                    self._debug(
                        f"ignoring keyword-only argument {repr(arg)} as it is set to "
                        f"its default value (set `argument_defaults` to change this "
                        f"behaviour)"
                    )
                    continue

                keyword[arg] = dump_param
            elif dump_param.kind is ArgumentType.ANY:
                # If there is a variadic positional argument, this has to be positional.
                # Otherwise, if the arg was previously parsed with a certain style, use
                # that. If not, use the preferred form. Furthermore, ignore default
                # values unless they were explicitly defined in the reference.
                if has_var_pos:
                    use_keyword = False
                elif arg in ref_args:
                    use_keyword = False
                elif arg in ref_kwargs:
                    use_keyword = True
                else:
                    use_keyword = prefer_keywords

                if use_keyword:
                    keyword[arg] = dump_param
                else:
                    positional[arg] = dump_param
            elif dump_param.kind is ArgumentType.VAR_POS:
                var_pos = dump_param
            elif dump_param.kind is ArgumentType.VAR_KEYWORD:
                var_keyword = dump_param
            else:
                raise RuntimeError(f"unhandled param kind {repr(dump_param.kind)}")

        if not self.argument_defaults:
            # Remove any trailing default positional arguments. Looping in reverse
            # order lets us remove defaults on the end but keep those between
            # non-defaults.
            for arg, dump_param in reversed(list(positional.items())):
                if arg not in ref_args and dump_param.is_default:
                    self._debug(
                        f"ignoring trailing positional argument {repr(arg)} as it "
                        f"is set to its default value (set `argument_defaults` to "
                        f"change this behaviour)"
                    )

                    del positional[arg]
                else:
                    # We've reached the first non-default positional argument.
                    break

            for kwarg, dump_param in list(keyword.items()):
                if kwarg not in ref_kwargs and dump_param.is_default:
                    self._debug(
                        f"ignoring keyword argument {repr(kwarg)} as it is set to "
                        f"its default value (set `argument_defaults` to change this "
                        f"behaviour)"
                    )

                    del keyword[kwarg]

        return positional, var_pos, keyword, var_keyword

    def _fill_delimited_args(
        self,
        positional,
        var_positional,
        keyword,
        var_keyword,
        path,
        delimiter_handler,
        ref_args=None,
        ref_kwargs=None,
        ref_extras=None,
    ):
        """Fill arguments, either generating new or reusing existing delimiters.

        If `ref_node` is given, the tokens at this node are reused if unchanged,
        otherwise those that have changed are generated anew with default formatting.
        """
        if ref_args is None:
            ref_args = {}
        if ref_kwargs is None:
            ref_kwargs = {}

        self._debug(
            f"Fill delimited argument inputs: {positional=}, {var_positional=}, "
            f"{keyword=}, {var_keyword=}, {ref_args=}, {ref_kwargs=}"
        )

        extra_tokens = []
        order = 0

        # Fill positional arguments.
        for name, value in positional.items():
            arg_ref_node = None
            arg_ref_extras = None
            first_arg_token = None

            if name in ref_args:
                # Reuse delimiter and extras up to the argument's first token.
                arg_ref_node, arg_ref_data = ref_args[name]
                arg_ref_extras = arg_ref_data.get("extra_tokens")
                first_arg_token = _first_instruction_token(arg_ref_data)

            # Reuse any extra tokens specified in the argument's parent up until
            # `token`, ensuring there is a delimiter.
            extra_tokens.extend(
                delimiter_handler(
                    order, until_token=first_arg_token, ref_extras=ref_extras
                )
            )

            self._fill_arg(value, order, path, arg_ref_node, arg_ref_extras)
            order += 1

        if var_positional:
            for value in var_positional.value:
                arg_ref_node = None
                arg_ref_extras = None

                # FIXME: implement reference dumping for variadic positional arguments.

                # Reuse any extra tokens specified in the argument's parent up until
                # `token`, ensuring there is a delimiter.
                extra_tokens.extend(delimiter_handler(order, ref_extras=ref_extras))

                self._fill_arg(value, order, path, arg_ref_node, arg_ref_extras)
                order += 1

        # Fill keyword arguments.
        for name, value in keyword.items():
            arg_ref_node = None
            arg_ref_extras = None
            first_arg_token = None

            if name in ref_kwargs:
                # Reuse delimiter and extras up to the argument's first token.
                arg_ref_node, arg_ref_data = ref_kwargs[name]
                arg_ref_extras = arg_ref_data.get("extra_tokens")
                first_arg_token = _first_instruction_token(arg_ref_data)

            # Reuse any extra tokens specified in the argument's parent up until
            # `token`, ensuring there is a delimiter.
            extra_tokens.extend(
                delimiter_handler(
                    order, until_token=first_arg_token, ref_extras=ref_extras
                )
            )

            self._fill_kwarg(name, value, order, path, arg_ref_node, arg_ref_extras)
            order += 1

        if var_keyword:
            for name, value in var_keyword.value.items():
                arg_ref_node = None
                arg_ref_extras = None

                # FIXME: implement reference dumping for variadic keyword arguments.

                # Reuse any extra tokens specified in the argument's parent up until
                # `token`, ensuring there is a delimiter.
                extra_tokens.extend(delimiter_handler(order, ref_extras=ref_extras))

                self._fill_kwarg(name, value, order, path, arg_ref_node, arg_ref_extras)
                order += 1

        return extra_tokens

    def _handle_script_arg_extras(self, order, until_token=None, ref_extras=None):
        """Yield element style extras, ensuring at least one delimiter is present,
        creating it if necessary."""
        has_delimiter = False

        if until_token and ref_extras:
            for extra in self._maybe_reuse_extras(ref_extras, until_token=until_token):
                if extra.type == "NEWLINE":
                    has_delimiter = True
                yield extra

        # Only prepend a delimiter if this isn't the first argument.
        if order > 0 and not has_delimiter:
            yield self._use_token(self._newline_token())

    def _delimit_element_arg(self, order, until_token=None, ref_extras=None):
        """Yield element style extras, ensuring at least one delimiter is present,
        creating it if necessary."""
        has_delimiter = False

        if until_token and ref_extras:
            for extra in self._maybe_reuse_extras(ref_extras, until_token=until_token):
                if extra.type == "WHITESPACE":
                    has_delimiter = True
                yield extra

        # Always prepend a delimiter because the whitespace between the name and first
        # argument is not added by the element filler.
        if not has_delimiter:
            yield self._use_token(self._space_token())

    def _delimit_function_arg(self, order, until_token=None, ref_extras=None):
        """Yield functional style extras, ensuring at least one delimiter is present,
        creating it if necessary."""
        has_delimiter = False

        if until_token and ref_extras:
            for extra in self._maybe_reuse_extras(ref_extras, until_token=until_token):
                if extra.type == "COMMA":
                    has_delimiter = True
                yield extra

        # Only prepend a delimiter if this isn't the first argument.
        if order > 0 and not has_delimiter:
            yield self._use_token(self._literal_token(","))
            yield self._use_token(self._space_token())

    def _fill_arg(self, value, order, path, ref_node, ref_extras):
        argument_path = self.graph.item_node_name(order, path)
        self._fill(value, argument_path, ref_node)
        self.graph.add_edge(argument_path, path, type=KatEdgeType.ARGUMENT, order=order)

    def _fill_kwarg(self, key, value, order, path, ref_node, ref_extras):
        if ref_node is not None:
            data = self.ref_graph.nodes[ref_node]

            # Reuse the key and '='.
            key_token = self._reuse_token(data["key_token"])
            equals_token = self._reuse_token(data["equals_token"])
        else:
            key_token = self._use_token(self._name_token(key))
            equals_token = self._use_token(self._literal_token("="))

        argument_path = self.graph.item_node_name(order, path)
        self._fill(
            value,
            argument_path,
            ref_node,
            key_token=key_token,
            equals_token=equals_token,
        )
        self.graph.add_edge(argument_path, path, type=KatEdgeType.ARGUMENT, order=order)


class KatUnfiller:
    """KatGraph to kat script."""

    def unfill(self, node, graph):
        fobj = StringIO()
        self.unfill_file(fobj, node, graph)
        fobj.seek(0)
        return fobj.read()

    def unfill_file(self, fobj, node, graph):
        production = self.production(node, graph)
        unparser = KatUnparser()
        unparser.unparse_file(fobj, production)

    def production(self, node, graph):
        data = graph.nodes[node]

        # Create any dependent arguments.
        arguments = scriptsorted(
            [
                self.production(argument_node, graph)
                for argument_node in graph.dependent_argument_nodes(node)
            ]
        )
        # Grab any extra tokens.
        extra = scriptsorted(data.get("extra_tokens", []))

        value = self._do_production(data["type"], data, arguments, extra)

        # Detect kwargs.
        if "key_token" in data:
            value = KatKwarg(
                key=data["key_token"], equals=data["equals_token"], value=value
            )

        return value

    @flagdispatchmethod
    def _do_production(self, nodetype, data, arguments, extra):
        raise NotImplementedError(f"unhandled KatGraph node type {repr(nodetype)}")

    @_do_production.register(KatNodeType.ROOT)
    def _(self, data, arguments, extra):
        return KatScript(arguments=arguments, extra=extra)

    @_do_production.register(KatNodeType.GENERATOR_TERMINAL_NODES)
    def _(self, data, arguments, extra):
        return data["token"]

    @_do_production.register(KatNodeType.ELEMENT)
    def _(self, data, arguments, extra):
        return KatElement(
            directive=data["token"],
            arguments=arguments,
            extra=extra,
            name=data.get("name_token"),
        )

    @_do_production.register(KatNodeType.FUNCTION)
    def _(self, data, arguments, extra):
        return KatFunction(directive=data["token"], arguments=arguments, extra=extra)

    @_do_production.register(KatNodeType.GROUPED_EXPRESSION)
    def _(self, data, arguments, extra):
        return KatGroupedExpression(arguments=arguments, extra=extra)

    @_do_production.register(KatNodeType.EXPRESSION)
    def _(self, data, arguments, extra):
        return KatExpression(operator=data["token"], arguments=arguments, extra=extra)

    @_do_production.register(KatNodeType.ARRAY)
    def _(self, data, arguments, extra):
        return KatArray(arguments=arguments, extra=extra)


class KatUnparser:
    """TokenContainer to kat script."""

    def unparse(self, container):
        fobj = StringIO()
        self.unparse_file(fobj, container)
        fobj.seek(0)
        return fobj.read()

    def unparse_file(self, fobj, container):
        untokenizer = KatUntokenizer()
        untokenizer.untokenize_file(fobj, container.sorted_tokens)


class KatUntokenizer:
    """Token to kat script."""

    def untokenize(self, tokens):
        fobj = StringIO()
        self.untokenize_file(fobj, tokens)
        fobj.seek(0)
        return fobj.read()

    def untokenize_file(self, fobj, tokens):
        for token in tokens:
            fobj.write(token.raw_value)
